from typing import Literal

from matplotlib.axes import Axes

from .event import view_event
from .frame import view_frame, view_frame_sequence
from .gif import export_frame_sequence_as_gif
from .png import export_frame_sequence_as_png


def view(
    obj: dict | list[dict],
    metadata: list[dict] | None = None,
    export: Literal['gif', 'png'] | None = None,
    ax: Axes | None = None,
):
    if export is None:
        if isinstance(obj, list):
            if len(obj) > 0:
                if 'homePlayersSmoothed' in obj[0] or 'awayPlayersSmoothed' in obj[0]:
                    return view_frame_sequence(obj, metadata, ax=ax)
                else:
                    ax = None
                    for evt in obj:
                        if ax is None:
                            ax = view_event(evt, ax=ax)
                        else:
                            view_event(evt, ax=ax)
        elif isinstance(obj, dict):
            if 'homePlayersSmoothed' in obj or 'awayPlayersSmoothed' in obj:
                print('HERE')
                return view_frame(obj, metadata, ax=ax)
            else:
                print('THERE')
                return view_event(obj, ax=ax)
    elif export == 'gif':
        if isinstance(obj, list):
            return export_frame_sequence_as_gif(obj, metadata)
        else:
            raise ValueError('GIF export only works with frame sequences')
    elif export == 'png':
        if isinstance(obj, list):
            return export_frame_sequence_as_png(obj, metadata)
    else:
        raise ValueError('Invalid export format. Choose between "gif" and "png"')



"""Jinja2 environment config for Gandula."""

from jinja2 import Environment, PackageLoader, select_autoescape

env = Environment(
    loader=PackageLoader('gandula', 'html_templates'), autoescape=select_autoescape()
)



"""PFF Event Validator.

This module provides functionality to validate PFF event data.
"""

from typing import Any

from gandula.schemas.pff_events import PFFEventGame


def validate_match(match: dict[str, Any]):
    """
    Validate a single PFF match dictionary.

    Args:
        match: Dictionary containing PFF match data

    Returns:
        Validated PFF match dictionary

    Raises:
        ValidationError: If validation fails
    """
    PFFEventGame.model_validate(match)



from typing import cast

from matplotlib.axes import Axes
from mplsoccer import Pitch


def view_event(event: dict, color='red', ax: Axes | None = None) -> Axes:
    pitch = Pitch(pitch_type='impect')
    if ax:
        pitch.draw(ax=ax)
    else:
        _, ax = pitch.draw()  # type: ignore
        ax = cast(Axes, ax)

    if event.get('end_x'):
        pitch.arrows(
            event['x'],
            event['y'],
            event['end_x'],
            event['end_y'],
            width=2,
            headwidth=5,
            headlength=5,
            color=color,
            ax=ax,
        )
    elif event.get('x'):
        pitch.scatter(event['x'], event['y'], color=color, ax=ax)

    return ax



import tempfile
from pathlib import Path

import orjson
import pandas as pd
from tqdm.auto import tqdm

from ..core.logging import get_logger
from ..io.reader import read_file, read_line
from .pff_gdrive_mapping import GDRIVE_FILE_IDS, PFF_MATCH_INFO

logger = get_logger(__name__)


def available() -> pd.DataFrame:
    return pd.DataFrame(PFF_MATCH_INFO)


def load_open_data(
    ids: int | list[int] | None = None,
    *,
    output_dir: str | None = None,
    events: bool = True,
    tracks: bool = True,
):
    if not events and not tracks:
        raise ValueError('At least one of events or tracks must be set to True')

    if isinstance(ids, int) and str(ids) not in GDRIVE_FILE_IDS:
        raise ValueError(f'Match {ids} not found')
    elif isinstance(ids, list) and all(str(_id) not in GDRIVE_FILE_IDS for _id in ids):
        raise ValueError(f'At least one match from {ids} was not found')

    if ids is None:
        logger.info('Loading all matches')
        matches = GDRIVE_FILE_IDS.keys()
    elif isinstance(ids, int):
        matches = [str(ids)]
    elif isinstance(ids, list):
        matches = [str(_id) for _id in ids]
    else:
        raise ValueError('Invalid match ID type')

    result = []
    for match_id in tqdm(matches, desc='Downloading matches'):
        try:
            data = _load_open_data_match(
                match_id,
                output_dir=output_dir,
                events=events,
                tracks=tracks,
            )
            result.append(data)
        except Exception as e:
            logger.exception(f'Error loading match {match_id}: {e}')
            continue

    if len(result) == 0:
        return []

    if len(result) == 1:
        return result[0]

    return result


def _load_open_data_match(
    id: str, *, output_dir: str | None = None, events=True, tracks=True
):
    output = None
    use_temp = output_dir is None

    if not use_temp:
        output = Path(output_dir)
        output.mkdir(parents=True, exist_ok=True)

    rosters_id = GDRIVE_FILE_IDS[id]['rosters']
    events_id = GDRIVE_FILE_IDS[id]['event']
    tracks_id = GDRIVE_FILE_IDS[id]['tracking']
    metadata_id = GDRIVE_FILE_IDS[id]['metadata']

    rosters = _load_from_gdrive(rosters_id, id, output, 'rosters')
    metadata = _load_from_gdrive(metadata_id, id, output, 'metadata')

    event_data = _load_from_gdrive(events_id, id, output, 'events') if events else None
    tracking_data = (
        _load_from_gdrive(tracks_id, id, output, 'tracks', track=True)
        if tracks
        else None
    )

    if events and tracks:
        return rosters, metadata, event_data, tracking_data

    return rosters, metadata, event_data or tracking_data


def _load_from_gdrive(
    id: str, match_id: str, output: Path | None, suffix: str | None, *, track=False
):
    if output is None:
        contents = None

        with tempfile.NamedTemporaryFile(delete_on_close=False) as fp:
            fp.close()

            import gdown

            gdown.download(id=id, output=fp.name, quiet=True)

            if track:
                contents = [orjson.loads(line) for line in read_line(fp.name, bz2=True)]
            else:
                contents = orjson.loads(read_file(fp.name))

        return contents

    if not suffix:
        output_file = output / f'{match_id}.json'
    else:
        output_file = output / f'{match_id}_{suffix}.json'

    import gdown

    gdown.download(id=id, output=str(output_file), quiet=True)
    return output_file



"""Complete field name mappings for PFF event data."""

EVENTS_FIELD_MAP = {
    # Game event fields
    'game_event_bodyType': 'game_event_body_type',
    'game_event_comment': 'game_event_comment',
    'game_event_duration': 'game_event_duration',
    'game_event_earlyDistribution': 'game_event_early_distribution',
    'game_event_endPointX': 'game_event_end_point_x',
    'game_event_endPointY': 'game_event_end_point_y',
    'game_event_endTime': 'game_event_end_time',
    'game_event_facingType': 'game_event_facing_type',
    'game_event_formattedGameClock': 'game_event_formatted_game_clock',
    'game_event_gameClock': 'game_event_game_clock',
    'game_event_gameEventType': 'game_event_type',
    'game_event_game_id': 'game_event_game_id',
    'game_event_id': 'game_event_id',
    'game_event_initialTouchType': 'game_event_initial_touch_type',
    'game_event_otherPlayer': 'game_event_other_player',
    'game_event_outType': 'game_event_out_type',
    'game_event_period': 'game_event_period',
    'game_event_player': 'game_event_player',
    'game_event_playerOff': 'game_event_player_off',
    'game_event_playerOffType': 'game_event_player_off_type',
    'game_event_playerOn': 'game_event_player_on',
    'game_event_player_id': 'game_event_player_id',
    'game_event_pressurePlayer': 'game_event_pressure_player',
    'game_event_pressurePlayer_id': 'game_event_pressure_player_id',
    'game_event_pressureType': 'game_event_pressure_type',
    'game_event_setpieceType': 'game_event_setpiece_type',
    'game_event_startPointX': 'game_event_start_point_x',
    'game_event_startPointY': 'game_event_start_point_y',
    'game_event_startTime': 'game_event_start_time',
    'game_event_team': 'game_event_team',
    'game_event_team_id': 'game_event_team_id',
    'game_event_team_name': 'game_event_team_name',
    'game_event_team_shortName': 'game_event_team_short_name',
    'game_event_touches': 'game_event_touches',
    'game_event_touchesInBox': 'game_event_touches_in_box',
    'game_event_videoUrl': 'game_event_video_url',
    # Possession event fields
    'possession_event_duration': 'possession_event_duration',
    'possession_event_endTime': 'possession_event_end_time',
    'possession_event_formattedGameClock': 'possession_event_formatted_game_clock',
    'possession_event_gameClock': 'possession_event_game_clock',
    'possession_event_gameEvent_id': 'possession_event_game_event_id',
    'possession_event_id': 'possession_event_id',
    'possession_event_insertedAt': 'possession_event_inserted_at',
    'possession_event_possessionEventType': 'possession_event_type',
    'possession_event_startTime': 'possession_event_start_time',
    'possession_event_updatedAt': 'possession_event_updated_at',
    'possession_event_videoUrl': 'possession_event_video_url',
    # Ball carry fields
    'ballCarry_additionalChallenger1': 'ball_carry_additional_challenger_1',
    'ballCarry_additionalChallenger2': 'ball_carry_additional_challenger_2',
    'ballCarry_additionalChallenger3': 'ball_carry_additional_challenger_3',
    'ballCarry_badParry': 'ball_carry_bad_parry',
    'ballCarry_ballCarrierPlayer_id': 'ball_carry_carrier_player_id',
    'ballCarry_ballCarryEndPointX': 'ball_carry_end_point_x',
    'ballCarry_ballCarryEndPointY': 'ball_carry_end_point_y',
    'ballCarry_ballCarryOutcome': 'ball_carry_outcome',
    'ballCarry_ballCarryStartPointX': 'ball_carry_start_point_x',
    'ballCarry_ballCarryStartPointY': 'ball_carry_start_point_y',
    'ballCarry_ballCarryType': 'ball_carry_type',
    'ballCarry_betterOption': 'ball_carry_better_option',
    'ballCarry_betterOptionPlayer': 'ball_carry_better_option_player',
    'ballCarry_betterOptionPlayer_id': 'ball_carry_better_option_player_id',
    'ballCarry_betterOptionTime': 'ball_carry_better_option_time',
    'ballCarry_betterOptionType': 'ball_carry_better_option_type',
    'ballCarry_carryIntent': 'ball_carry_intent',
    'ballCarry_carryType': 'ball_carry_type',
    'ballCarry_challengerPlayer': 'ball_carry_challenger_player',
    'ballCarry_clearerPlayer': 'ball_carry_clearer_player',
    'ballCarry_createsSpace': 'ball_carry_creates_space',
    'ballCarry_defenderPlayer': 'ball_carry_defender_player',
    'ballCarry_defenderPlayer_id': 'ball_carry_defender_player_id',
    'ballCarry_defenderPointX': 'ball_carry_defender_point_x',
    'ballCarry_defenderPointY': 'ball_carry_defender_point_y',
    'ballCarry_dribbleEndPointX': 'ball_carry_dribble_end_point_x',
    'ballCarry_dribbleEndPointY': 'ball_carry_dribble_end_point_y',
    'ballCarry_dribbleOutcomeType': 'ball_carry_dribble_outcome_type',
    'ballCarry_dribbleStartPointX': 'ball_carry_dribble_start_point_x',
    'ballCarry_dribbleStartPointY': 'ball_carry_dribble_start_point_y',
    'ballCarry_dribbleType': 'ball_carry_dribble_type',
    'ballCarry_gameEvent_id': 'ball_carry_game_event_id',
    'ballCarry_game_id': 'ball_carry_game_id',
    'ballCarry_id': 'ball_carry_id',
    'ballCarry_keeperPlayer': 'ball_carry_keeper_player',
    'ballCarry_late': 'ball_carry_late',
    'ballCarry_linesBrokenType': 'ball_carry_lines_broken_type',
    'ballCarry_missedTouchPlayer': 'ball_carry_missed_touch_player',
    'ballCarry_missedTouchPointX': 'ball_carry_missed_touch_point_x',
    'ballCarry_missedTouchPointY': 'ball_carry_missed_touch_point_y',
    'ballCarry_missedTouchType': 'ball_carry_missed_touch_type',
    'ballCarry_opportunityType': 'ball_carry_opportunity_type',
    'ballCarry_period': 'ball_carry_period',
    'ballCarry_possessionEvent_id': 'ball_carry_possession_event_id',
    'ballCarry_pressurePlayer': 'ball_carry_pressure_player',
    'ballCarry_pressurePlayer_id': 'ball_carry_pressure_player_id',
    'ballCarry_pressureType': 'ball_carry_pressure_type',
    'ballCarry_saveReboundType': 'ball_carry_save_rebound_type',
    'ballCarry_saveable': 'ball_carry_saveable',
    'ballCarry_shotInitialHeightType': 'ball_carry_shot_initial_height_type',
    'ballCarry_shotOutcomeType': 'ball_carry_shot_outcome_type',
    'ballCarry_successful': 'ball_carry_successful',
    'ballCarry_tackleAttemptPointX': 'ball_carry_tackle_attempt_point_x',
    'ballCarry_tackleAttemptPointY': 'ball_carry_tackle_attempt_point_y',
    'ballCarry_tackleAttemptType': 'ball_carry_tackle_attempt_type',
    'ballCarry_touchOutcomePlayer': 'ball_carry_touch_outcome_player',
    'ballCarry_touchOutcomeType': 'ball_carry_touch_outcome_type',
    'ballCarry_touchPointX': 'ball_carry_touch_point_x',
    'ballCarry_touchPointY': 'ball_carry_touch_point_y',
    'ballCarry_touchType': 'ball_carry_touch_type',
    'ballCarry_trickType': 'ball_carry_trick_type',
    # Challenge fields
    'challenge_additionalChallenger1': 'challenge_additional_challenger_1',
    'challenge_additionalChallenger2': 'challenge_additional_challenger_2',
    'challenge_additionalChallenger3': 'challenge_additional_challenger_3',
    'challenge_advantageType': 'challenge_advantage_type',
    'challenge_ballCarrierPlayer': 'challenge_ball_carrier_player',
    'challenge_ballCarrierPlayer_id': 'challenge_ball_carrier_player_id',
    'challenge_carryType': 'challenge_carry_type',
    'challenge_challengeOutcomeType': 'challenge_outcome_type',
    'challenge_challengePointX': 'challenge_point_x',
    'challenge_challengePointY': 'challenge_point_y',
    'challenge_challengeType': 'challenge_type',
    'challenge_challengeWinnerPlayer': 'challenge_winner_player',
    'challenge_challengeWinnerPlayer_id': 'challenge_winner_player_id',
    'challenge_challengerAwayPlayer': 'challenge_challenger_away_player',
    'challenge_challengerAwayPlayer_id': 'challenge_challenger_away_player_id',
    'challenge_challengerHomePlayer': 'challenge_challenger_home_player',
    'challenge_challengerHomePlayer_id': 'challenge_challenger_home_player_id',
    'challenge_challengerPlayer': 'challenge_challenger_player',
    'challenge_challengerPlayer_id': 'challenge_challenger_player_id',
    'challenge_createsSpace': 'challenge_creates_space',
    'challenge_dribbleEndPointX': 'challenge_dribble_end_point_x',
    'challenge_dribbleEndPointY': 'challenge_dribble_end_point_y',
    'challenge_dribbleStartPointX': 'challenge_dribble_start_point_x',
    'challenge_dribbleStartPointY': 'challenge_dribble_start_point_y',
    'challenge_dribbleType': 'challenge_dribble_type',
    'challenge_gameEvent_id': 'challenge_game_event_id',
    'challenge_game_id': 'challenge_game_id',
    'challenge_id': 'challenge_id',
    'challenge_insertedAt': 'challenge_inserted_at',
    'challenge_keeperPlayer': 'challenge_keeper_player',
    'challenge_linesBrokenType': 'challenge_lines_broken_type',
    'challenge_missedTouchPlayer': 'challenge_missed_touch_player',
    'challenge_missedTouchPointX': 'challenge_missed_touch_point_x',
    'challenge_missedTouchPointY': 'challenge_missed_touch_point_y',
    'challenge_missedTouchType': 'challenge_missed_touch_type',
    'challenge_opportunityType': 'challenge_opportunity_type',
    'challenge_period': 'challenge_period',
    'challenge_possessionEvent_id': 'challenge_possession_event_id',
    'challenge_pressurePlayer': 'challenge_pressure_player',
    'challenge_tackleAttemptPointX': 'challenge_tackle_attempt_point_x',
    'challenge_tackleAttemptPointY': 'challenge_tackle_attempt_point_y',
    'challenge_tackleAttemptType': 'challenge_tackle_attempt_type',
    'challenge_trickType': 'challenge_trick_type',
    # Clearance fields
    'clearance_advantageType': 'clearance_advantage_type',
    'clearance_badParry': 'clearance_bad_parry',
    'clearance_ballHeightType': 'clearance_ball_height_type',
    'clearance_ballHighPointType': 'clearance_ball_high_point_type',
    'clearance_betterOption': 'clearance_better_option',
    'clearance_betterOptionPlayer': 'clearance_better_option_player',
    'clearance_betterOptionPlayer_id': 'clearance_better_option_player_id',
    'clearance_betterOptionTime': 'clearance_better_option_time',
    'clearance_betterOptionType': 'clearance_better_option_type',
    'clearance_blockerPlayer': 'clearance_blocker_player',
    'clearance_blockerPlayer_id': 'clearance_blocker_player_id',
    'clearance_clearanceBodyType': 'clearance_body_type',
    'clearance_clearanceEndPointX': 'clearance_end_point_x',
    'clearance_clearanceEndPointY': 'clearance_end_point_y',
    'clearance_clearanceOutcomeType': 'clearance_outcome_type',
    'clearance_clearancePlayer_id': 'clearance_player_id',
    'clearance_clearancePointX': 'clearance_point_x',
    'clearance_clearancePointY': 'clearance_point_y',
    'clearance_clearanceStartPointX': 'clearance_start_point_x',
    'clearance_clearanceStartPointY': 'clearance_start_point_y',
    'clearance_clearerPlayer': 'clearance_clearer_player',
    'clearance_createsSpace': 'clearance_creates_space',
    'clearance_createsSpaceVsPlayer': 'clearance_creates_space_vs_player',
    'clearance_defenderPlayer': 'clearance_defender_player',
    'clearance_failedInterventionPlayer': 'clearance_failed_intervention_player',
    'clearance_failedInterventionPlayer1': 'clearance_failed_intervention_player_1',
    'clearance_failedInterventionPlayer2': 'clearance_failed_intervention_player_2',
    'clearance_failedInterventionPlayer3': 'clearance_failed_intervention_player_3',
    'clearance_failedInterventionPlayer_id': 'clearance_failed_intervention_player_id',
    'clearance_gameEvent_id': 'clearance_game_event_id',
    'clearance_game_id': 'clearance_game_id',
    'clearance_heightType': 'clearance_height_type',
    'clearance_id': 'clearance_id',
    'clearance_keeperPlayer': 'clearance_keeper_player',
    'clearance_late': 'clearance_late',
    'clearance_missedTouchPlayer': 'clearance_missed_touch_player',
    'clearance_missedTouchPointX': 'clearance_missed_touch_point_x',
    'clearance_missedTouchPointY': 'clearance_missed_touch_point_y',
    'clearance_missedTouchType': 'clearance_missed_touch_type',
    'clearance_opportunityType': 'clearance_opportunity_type',
    'clearance_period': 'clearance_period',
    'clearance_possessionEvent_id': 'clearance_possession_event_id',
    'clearance_pressurePlayer': 'clearance_pressure_player',
    'clearance_pressurePlayer_id': 'clearance_pressure_player_id',
    'clearance_pressureType': 'clearance_pressure_type',
    'clearance_saveReboundType': 'clearance_save_rebound_type',
    'clearance_saveable': 'clearance_saveable',
    'clearance_shotInitialHeightType': 'clearance_shot_initial_height_type',
    'clearance_shotOutcomeType': 'clearance_shot_outcome_type',
    # Cross fields
    'cross_advantageType': 'cross_advantage_type',
    'cross_badParry': 'cross_bad_parry',
    'cross_ballHeightType': 'cross_ball_height_type',
    'cross_betterOption': 'cross_better_option',
    'cross_betterOptionPlayer': 'cross_better_option_player',
    'cross_betterOptionTime': 'cross_better_option_time',
    'cross_betterOptionType': 'cross_better_option_type',
    'cross_blockerPlayer': 'cross_blocker_player',
    'cross_blockerPlayer_id': 'cross_blocker_player_id',
    'cross_clearerPlayer': 'cross_clearer_player',
    'cross_completeToPlayer': 'cross_complete_to_player',
    'cross_completeToPlayer_id': 'cross_complete_to_player_id',
    'cross_createsSpace': 'cross_creates_space',
    'cross_createsSpaceVsPlayer': 'cross_creates_space_vs_player',
    'cross_createsSpaceVsPlayer_id': 'cross_creates_space_vs_player_id',
    'cross_crossAccuracyType': 'cross_accuracy_type',
    'cross_crossEndPointX': 'cross_end_point_x',
    'cross_crossEndPointY': 'cross_end_point_y',
    'cross_crossHighPointType': 'cross_high_point_type',
    'cross_crossOutcomeType': 'cross_outcome_type',
    'cross_crossPointX': 'cross_point_x',
    'cross_crossPointY': 'cross_point_y',
    'cross_crossStartPointX': 'cross_start_point_x',
    'cross_crossStartPointY': 'cross_start_point_y',
    'cross_crossType': 'cross_type',
    'cross_crossZoneType': 'cross_zone_type',
    'cross_crosserBodyType': 'cross_body_type',
    'cross_crosserPlayer_id': 'cross_player_id',
    'cross_defenderBallHeightType': 'cross_defender_ball_height_type',
    'cross_defenderBodyType': 'cross_defender_body_type',
    'cross_defenderPlayer': 'cross_defender_player',
    'cross_defenderPlayer_id': 'cross_defender_player_id',
    'cross_deflectionPointX': 'cross_deflection_point_x',
    'cross_deflectionPointY': 'cross_deflection_point_y',
    'cross_deflectorBodyType': 'cross_deflector_body_type',
    'cross_deflectorPlayer': 'cross_deflector_player',
    'cross_deflectorPlayer_id': 'cross_deflector_player_id',
    'cross_failedInterventionPlayer': 'cross_failed_intervention_player',
    'cross_failedInterventionPlayer1': 'cross_failed_intervention_player_1',
    'cross_failedInterventionPlayer2': 'cross_failed_intervention_player_2',
    'cross_failedInterventionPlayer3': 'cross_failed_intervention_player_3',
    'cross_failedInterventionPlayer_id': 'cross_failed_intervention_player_id',
    'cross_failedInterventionPointX': 'cross_failed_intervention_point_x',
    'cross_failedInterventionPointY': 'cross_failed_intervention_point_y',
    'cross_gameEvent_id': 'cross_game_event_id',
    'cross_game_id': 'cross_game_id',
    'cross_goalkeeperPointX': 'cross_goalkeeper_point_x',
    'cross_goalkeeperPointY': 'cross_goalkeeper_point_y',
    'cross_id': 'cross_id',
    'cross_incompletionReasonType': 'cross_incompletion_reason_type',
    'cross_intendedTargetPlayer': 'cross_intended_target_player',
    'cross_intendedTargetPlayer_id': 'cross_intended_target_player_id',
    'cross_intendedTargetPointX': 'cross_intended_target_point_x',
    'cross_intendedTargetPointY': 'cross_intended_target_point_y',
    'cross_keeperPlayer': 'cross_keeper_player',
    'cross_late': 'cross_late',
    'cross_linesBrokenType': 'cross_lines_broken_type',
    'cross_missedTouchPlayer': 'cross_missed_touch_player',
    'cross_missedTouchPlayer_id': 'cross_missed_touch_player_id',
    'cross_missedTouchPointX': 'cross_missed_touch_point_x',
    'cross_missedTouchPointY': 'cross_missed_touch_point_y',
    'cross_missedTouchType': 'cross_missed_touch_type',
    'cross_noLook': 'cross_no_look',
    'cross_opportunityType': 'cross_opportunity_type',
    'cross_period': 'cross_period',
    'cross_possessionEvent_id': 'cross_possession_event_id',
    'cross_pressurePlayer': 'cross_pressure_player',
    'cross_pressurePlayer_id': 'cross_pressure_player_id',
    'cross_pressureType': 'cross_pressure_type',
    'cross_receiverBallHeightType': 'cross_receiver_ball_height_type',
    'cross_receiverBodyType': 'cross_receiver_body_type',
    'cross_receiverPointX': 'cross_receiver_point_x',
    'cross_receiverPointY': 'cross_receiver_point_y',
    'cross_saveReboundType': 'cross_save_rebound_type',
    'cross_saveable': 'cross_saveable',
    'cross_secondIncompletionReasonType': 'cross_second_incompletion_reason_type',
    'cross_shotInitialHeightType': 'cross_shot_initial_height_type',
    'cross_shotOutcomeType': 'cross_shot_outcome_type',
    'cross_targetZonePointX': 'cross_target_zone_point_x',
    'cross_targetZonePointY': 'cross_target_zone_point_y',
    # Foul fields
    'foul_badCall': 'foul_bad_call',
    'foul_correctDecision': 'foul_correct_decision',
    'foul_culpritPlayer_id': 'foul_culprit_player_id',
    'foul_foulOutcomeType': 'foul_outcome_type',
    'foul_foulPointX': 'foul_point_x',
    'foul_foulPointY': 'foul_point_y',
    'foul_foulType': 'foul_type',
    'foul_gameEvent_id': 'foul_game_event_id',
    'foul_game_id': 'foul_game_id',
    'foul_id': 'foul_id',
    'foul_possessionEvent_id': 'foul_possession_event_id',
    'foul_potentialOffenseType': 'foul_potential_offense_type',
    'foul_sequence': 'foul_sequence',
    'foul_tacticalFoul': 'foul_tactical',
    'foul_var': 'foul_var',
    'foul_varCulpritPlayer': 'foul_var_culprit_player',
    'foul_varOutcomeType': 'foul_var_outcome_type',
    'foul_varPotentialOffenseType': 'foul_var_potential_offense_type',
    'foul_varReasonType': 'foul_var_reason_type',
    'foul_victimPlayer': 'foul_victim_player',
    'foul_victimPlayer_id': 'foul_victim_player_id',
    # Passing fields
    'passing_advantageType': 'passing_advantage_type',
    'passing_badParry': 'passing_bad_parry',
    'passing_ballHeightType': 'passing_ball_height_type',
    'passing_betterOption': 'passing_better_option',
    'passing_betterOptionPlayer': 'passing_better_option_player',
    'passing_betterOptionPlayer_id': 'passing_better_option_player_id',
    'passing_betterOptionTime': 'passing_better_option_time',
    'passing_betterOptionType': 'passing_better_option_type',
    'passing_blockerPlayer': 'passing_blocker_player',
    'passing_blockerPlayer_id': 'passing_blocker_player_id',
    'passing_clearerPlayer': 'passing_clearer_player',
    'passing_createsSpace': 'passing_creates_space',
    'passing_createsSpaceVsPlayer': 'passing_creates_space_vs_player',
    'passing_createsSpaceVsPlayer_id': 'passing_creates_space_vs_player_id',
    'passing_defenderBodyType': 'passing_defender_body_type',
    'passing_defenderHeightType': 'passing_defender_height_type',
    'passing_defenderPlayer': 'passing_defender_player',
    'passing_defenderPlayer_id': 'passing_defender_player_id',
    'passing_defenderPointX': 'passing_defender_point_x',
    'passing_defenderPointY': 'passing_defender_point_y',
    'passing_deflectionPointX': 'passing_deflection_point_x',
    'passing_deflectionPointY': 'passing_deflection_point_y',
    'passing_deflectorBodyType': 'passing_deflector_body_type',
    'passing_deflectorPlayer': 'passing_deflector_player',
    'passing_deflectorPlayer_id': 'passing_deflector_player_id',
    'passing_failedInterventionPlayer': 'passing_failed_intervention_player',
    'passing_failedInterventionPlayer1': 'passing_failed_intervention_player_1',
    'passing_failedInterventionPlayer2': 'passing_failed_intervention_player_2',
    'passing_failedInterventionPlayer3': 'passing_failed_intervention_player_3',
    'passing_failedInterventionPlayer_id': 'passing_failed_intervention_player_id',
    'passing_failedInterventionPointX': 'passing_failed_intervention_point_x',
    'passing_failedInterventionPointY': 'passing_failed_intervention_point_y',
    'passing_gameEvent_id': 'passing_game_event_id',
    'passing_game_id': 'passing_game_id',
    'passing_goalkeeperPointX': 'passing_goalkeeper_point_x',
    'passing_goalkeeperPointY': 'passing_goalkeeper_point_y',
    'passing_id': 'passing_id',
    'passing_incompletionReasonType': 'passing_incompletion_reason_type',
    'passing_keeperPlayer': 'passing_keeper_player',
    'passing_late': 'passing_late',
    'passing_linesBrokenType': 'passing_lines_broken_type',
    'passing_missedTouchPlayer': 'passing_missed_touch_player',
    'passing_missedTouchPlayer_id': 'passing_missed_touch_player_id',
    'passing_missedTouchPointX': 'passing_missed_touch_point_x',
    'passing_missedTouchPointY': 'passing_missed_touch_point_y',
    'passing_missedTouchType': 'passing_missed_touch_type',
    'passing_noLook': 'passing_no_look',
    'passing_onTarget': 'passing_on_target',
    'passing_opportunityType': 'passing_opportunity_type',
    'passing_outOfPlayPointX': 'passing_out_of_play_point_x',
    'passing_outOfPlayPointY': 'passing_out_of_play_point_y',
    'passing_passerPlayer_id': 'passing_passer_player_id',
    'passing_passBodyType': 'passing_body_type',
    'passing_passHighPointType': 'passing_high_point_type',
    'passing_passOutcomeType': 'passing_outcome_type',
    'passing_passPointX': 'passing_point_x',
    'passing_passPointY': 'passing_point_y',
    'passing_passType': 'passing_type',
    'passing_passAccuracyType': 'passing_accuracy_type',
    'passing_period': 'passing_period',
    'passing_possessionEvent_id': 'passing_possession_event_id',
    'passing_pressurePlayer': 'passing_pressure_player',
    'passing_pressurePlayer_id': 'passing_pressure_player_id',
    'passing_pressureType': 'passing_pressure_type',
    'passing_receiverBodyType': 'passing_receiver_body_type',
    'passing_receiverFacingType': 'passing_receiver_facing_type',
    'passing_receiverHeightType': 'passing_receiver_height_type',
    'passing_receiverPlayer': 'passing_receiver_player',
    'passing_receiverPlayer_id': 'passing_receiver_player_id',
    'passing_receiverPointX': 'passing_receiver_point_x',
    'passing_receiverPointY': 'passing_receiver_point_y',
    'passing_saveReboundType': 'passing_save_rebound_type',
    'passing_saveable': 'passing_saveable',
    'passing_secondIncompletionReasonType': 'passing_second_incompletion_reason_type',
    'passing_shotInitialHeightType': 'passing_shot_initial_height_type',
    'passing_shotOutcomeType': 'passing_shot_outcome_type',
    'passing_targetFacingType': 'passing_target_facing_type',
    'passing_targetPlayer_id': 'passing_target_player_id',
    'passing_targetPointX': 'passing_target_point_x',
    'passing_targetPointY': 'passing_target_point_y',
    # Rebound fields
    'rebound_advantageType': 'rebound_advantage_type',
    'rebound_badParry': 'rebound_bad_parry',
    'rebound_blockerPlayer': 'rebound_blocker_player',
    'rebound_defenderPlayer': 'rebound_defender_player',
    'rebound_gameEvent_id': 'rebound_game_event_id',
    'rebound_game_id': 'rebound_game_id',
    'rebound_id': 'rebound_id',
    'rebound_keeperPlayer': 'rebound_keeper_player',
    'rebound_missedTouchPlayer': 'rebound_missed_touch_player',
    'rebound_missedTouchPointX': 'rebound_missed_touch_point_x',
    'rebound_missedTouchPointY': 'rebound_missed_touch_point_y',
    'rebound_missedTouchType': 'rebound_missed_touch_type',
    'rebound_originateType': 'rebound_originate_type',
    'rebound_period': 'rebound_period',
    'rebound_possessionEvent_id': 'rebound_possession_event_id',
    'rebound_reboundBodyType': 'rebound_body_type',
    'rebound_reboundEndPointX': 'rebound_end_point_x',
    'rebound_reboundEndPointY': 'rebound_end_point_y',
    'rebound_reboundHeightType': 'rebound_height_type',
    'rebound_reboundHighPointType': 'rebound_high_point_type',
    'rebound_reboundOutcomeType': 'rebound_outcome_type',
    'rebound_reboundPointX': 'rebound_point_x',
    'rebound_reboundPointY': 'rebound_point_y',
    'rebound_reboundStartPointX': 'rebound_start_point_x',
    'rebound_reboundStartPointY': 'rebound_start_point_y',
    'rebound_rebounderPlayer_id': 'rebound_player_id',
    'rebound_saveReboundType': 'rebound_save_rebound_type',
    'rebound_saveable': 'rebound_saveable',
    'rebound_shotInitialHeightType': 'rebound_shot_initial_height_type',
    'rebound_shotOutcomeType': 'rebound_shot_outcome_type',
    # Shooting fields
    'shooting_advantageType': 'shooting_advantage_type',
    'shooting_badParry': 'shooting_bad_parry',
    'shooting_ballHeightType': 'shooting_ball_height_type',
    'shooting_ballMoving': 'shooting_ball_moving',
    'shooting_betterOption': 'shooting_better_option',
    'shooting_betterOptionPlayer': 'shooting_better_option_player',
    'shooting_betterOptionPlayer_id': 'shooting_better_option_player_id',
    'shooting_betterOptionTime': 'shooting_better_option_time',
    'shooting_betterOptionType': 'shooting_better_option_type',
    'shooting_blockerPlayer': 'shooting_blocker_player',
    'shooting_blockerPlayer_id': 'shooting_blocker_player_id',
    'shooting_bodyMovementType': 'shooting_body_movement_type',
    'shooting_clearerPlayer': 'shooting_clearer_player',
    'shooting_createsSpace': 'shooting_creates_space',
    'shooting_createsSpaceVsPlayer': 'shooting_creates_space_vs_player',
    'shooting_createsSpaceVsPlayer_id': 'shooting_creates_space_vs_player_id',
    'shooting_defenderPointX': 'shooting_defender_point_x',
    'shooting_defenderPointY': 'shooting_defender_point_y',
    'shooting_deflectionPointX': 'shooting_deflection_point_x',
    'shooting_deflectionPointY': 'shooting_deflection_point_y',
    'shooting_deflectorBodyType': 'shooting_deflector_body_type',
    'shooting_deflectorPlayer': 'shooting_deflector_player',
    'shooting_failedInterventionPlayer': 'shooting_failed_intervention_player',
    'shooting_failedInterventionPlayer1': 'shooting_failed_intervention_player_1',
    'shooting_failedInterventionPlayer2': 'shooting_failed_intervention_player_2',
    'shooting_failedInterventionPlayer3': 'shooting_failed_intervention_player_3',
    'shooting_failedInterventionPlayer_id': 'shooting_failed_intervention_player_id',
    'shooting_gameEvent_id': 'shooting_game_event_id',
    'shooting_game_id': 'shooting_game_id',
    'shooting_goalLineEndPointX': 'shooting_goal_line_end_point_x',
    'shooting_goalLineEndPointY': 'shooting_goal_line_end_point_y',
    'shooting_goalLineStartPointX': 'shooting_goal_line_start_point_x',
    'shooting_goalLineStartPointY': 'shooting_goal_line_start_point_y',
    'shooting_goalkeeperPointX': 'shooting_goalkeeper_point_x',
    'shooting_goalkeeperPointY': 'shooting_goalkeeper_point_y',
    'shooting_id': 'shooting_id',
    'shooting_keeperTouchPointX': 'shooting_keeper_touch_point_x',
    'shooting_keeperTouchPointY': 'shooting_keeper_touch_point_y',
    'shooting_keeperTouchType': 'shooting_keeper_touch_type',
    'shooting_late': 'shooting_late',
    'shooting_missedTouchPlayer': 'shooting_missed_touch_player',
    'shooting_missedTouchPointX': 'shooting_missed_touch_point_x',
    'shooting_missedTouchPointY': 'shooting_missed_touch_point_y',
    'shooting_missedTouchType': 'shooting_missed_touch_type',
    'shooting_noLook': 'shooting_no_look',
    'shooting_period': 'shooting_period',
    'shooting_possessionEvent_id': 'shooting_possession_event_id',
    'shooting_pressurePlayer': 'shooting_pressure_player',
    'shooting_pressurePlayer_id': 'shooting_pressure_player_id',
    'shooting_pressureType': 'shooting_pressure_type',
    'shooting_saveHeightType': 'shooting_save_height_type',
    'shooting_savePointX': 'shooting_save_point_x',
    'shooting_savePointY': 'shooting_save_point_y',
    'shooting_saveReboundType': 'shooting_save_rebound_type',
    'shooting_saveable': 'shooting_saveable',
    'shooting_saverPlayer': 'shooting_saver_player',
    'shooting_saverPlayer_id': 'shooting_saver_player_id',
    'shooting_shooterPlayer_id': 'shooting_shooter_player_id',
    'shooting_shotBodyType': 'shooting_body_type',
    'shooting_shotInitialHeightType': 'shooting_initial_height_type',
    'shooting_shotNatureType': 'shooting_nature_type',
    'shooting_shotOutcomeType': 'shooting_outcome_type',
    'shooting_shotPointX': 'shooting_point_x',
    'shooting_shotPointY': 'shooting_point_y',
    'shooting_shotTargetPointX': 'shooting_target_point_x',
    'shooting_shotTargetPointY': 'shooting_target_point_y',
    'shooting_shotType': 'shooting_type',
}



# Only possession event types with start_x, start_y, and maybe end_x, end_y.
POSSESSION_EVENT_TYPES_NEW_FORMAT = {
    'BALL_CARRY': 'BC',
    'PASS': 'PA',
    'SHOT': 'SH',
    'CHALLENGE': 'CH',
    'CROSS': 'CR',
    'CLEARANCE': 'CL',
}


def enhance_pff(events: list[dict], tracks) -> list[dict]:
    for event in events:
        try:
            add_starting_coordinates_from_freeze_frame(event)
            frame = find_event_end_frame(event, tracks)
            if frame:
                add_end_coordinates_from_frame(event, frame)
        except ValueError:
            continue
    return events


def add_starting_coordinates_from_freeze_frame(event: dict) -> dict:
    player_id = event['gameEvents']['playerId']
    for player in event['homePlayers'] + event['awayPlayers']:
        if player['playerId'] == player_id:
            event['x'] = player['x']
            event['y'] = player['y']
            return event
    # event['x'] = None
    # event['y'] = None
    # return event
    raise ValueError('Player not found in event')


def add_end_coordinates_from_frame(event: dict, frame: dict) -> dict:
    possession_event_type = event['possessionEvents']['possessionEventType']
    if possession_event_type is None:
        return event

    player_shirt = frame['game_event']['shirt_number']
    player = next(
        (
            player
            for player in frame['homePlayers'] + frame['awayPlayers']
            if player['jerseyNum'] == player_shirt
        ),
        None,
    )
    if player is None:
        return event

    event['end_x'] = player['x']
    event['end_y'] = player['y']
    return event


def find_event_end_frame(event: dict, tracks: list[dict]) -> dict | None:
    possession_event_type = event['possessionEvents']['possessionEventType']

    if possession_event_type == 'BC':
        frame = next(
            (
                track
                for track in tracks
                if track['possession_event_id'] == event['possessionEventId']
            ),
            None,
        )
        return frame
    elif possession_event_type in ['CL', 'CR', 'PA']:
        iterator = iter(tracks)
        _ = next(
            (
                track
                for track in iterator
                if track['possession_event_id'] == event['possessionEventId']
            ),
            None,
        )
        next_frame = next(
            (
                track
                for track in iterator
                if track['possession_event'] is not None
                and track['game_event_id'] is not None
                and track['possession_event']['possession_event_type'] != 'RE'
            ),
            None,
        )
        return next_frame



import time
from pathlib import Path

from matplotlib.pyplot import close, savefig

from .frame import view_frame


def export_frame_sequence_as_png(
    frames: dict | list[dict],
    metadata: list[dict] | None = None,
    filename: str | None = None,
) -> list[str]:
    paths = []

    if isinstance(frames, dict):
        frames = [frames]

    if filename is None:
        filename = f'gandula_img_{int(time.time() * 1000)}'

    if filename.endswith('.png'):
        filename = filename[:-4]

    for index, frame in enumerate(frames):
        ax = view_frame(frame, metadata)
        fig = ax.get_figure()
        filepath = f'{filename}_{index}.png'
        savefig(filepath, bbox_inches='tight', dpi=100)
        paths.append(str(Path(filepath).resolve()))
        close(fig)

    return paths



"""
Core functionality for gandula.

Provides configuration and logging capabilities.
"""

from .config import configure, get_config_value, get_data_dir, get_temp_dir
from .logging import configure_logging, get_logger

__all__ = [
    'configure',
    'get_config_value',
    'get_data_dir',
    'get_temp_dir',
    'configure_logging',
    'get_logger',
]



"""PFF Track Validator.

This module provides functionality to validate PFF tracking data stored in both regular
and compressed JSONL files.
"""

from collections.abc import Iterator
from pathlib import Path
from time import sleep

from tqdm import tqdm

from ..core.logging import get_logger
from ..io.reader import find_track_files, read_file
from ..schemas.pff_tracks import PFFFrame

logger = get_logger(__name__)


def validate_records(records: Iterator[str]):
    """
    Validate PFF track records against the PFFFrame schema.

    Args:
        records: Iterator of JSON strings to validate

    Returns:
        List of validated PFFFrame objects

    Raises:
        ValidationError: If validation fails for any record
    """
    [PFFFrame.model_validate_json(record) for record in records]


def validate_file(file_path: Path):
    """
    Validate a single PFF track file.

    Args:
        file_path: Path to the jsonl or jsonl.bz2 file

    Returns:
        List of validated PFFFrame objects

    Raises:
        ValueError: If the file cannot be read or processed
    """
    if not file_path.exists():
        raise ValueError(f'File does not exist: {file_path}')
    records = read_file(file_path)
    validate_records(records)


def validate_directory(path: Path):
    """
    Validate all PFF track files in a directory.

    Args:
        path: Path to the directory containing JSONL and JSONL.BZ2 files

    Returns:
        Dictionary mapping filenames to either lists of validated PFFFrame objects
        or exceptions if processing the file failed

    Raises:
        ValueError: If the path does not exist or is not a directory
    """
    if not path.exists():
        raise ValueError(f'Directory does not exist: {path}')
    if not path.is_dir():
        raise ValueError(f'Path is not a directory: {path}')

    files = find_track_files(path)

    with tqdm(files, total=len(files), desc='Validating tracks') as pbar:
        for file_path in files:
            records = read_file(file_path)
            validate_records(records)
            sleep(0.1)
            pbar.update(1)



from .enrich.pff import enhance_pff
from .providers.pff import available, load_open_data
from .view import view
from .view.video import video

__all__ = ['available', 'load_open_data', 'enhance_pff', 'view', 'video']



from datetime import timedelta

from kloppy.domain import (
    Period,
    Point,
    Point3D,
    Team,
)
from kloppy.exceptions import DeserializationError


def get_team_by_id(team_id: int, teams: list[Team]) -> Team:
    """Get a team by its id."""
    if str(team_id) == teams[0].team_id:
        return teams[0]
    elif str(team_id) == teams[1].team_id:
        return teams[1]
    else:
        raise DeserializationError(f'Unknown team_id {team_id}')


def get_team_by_player_id(player_id: int | str, teams: list[Team]) -> Team:
    """Get a team by a player's id."""
    for team in teams:
        for player in team.players:
            if player.player_id == str(player_id):
                return team
    raise DeserializationError(f'Unknown player_id {player_id}')


def get_period_by_id(period_id: int | str, periods: list[Period]) -> Period:
    """Get a period by its id."""
    for period in periods:
        if period.id == int(period_id):
            return period
    raise DeserializationError(f'Unknown period_id {period_id}')


from ..core.logging import get_logger

logger = get_logger(__name__)


def get_period_by_timestamp(timestamp: timedelta, periods: list[Period]) -> Period:
    """Get a period by its timestamp."""
    for period in periods:
        if period.start_timestamp <= timestamp <= period.end_timestamp:
            return period
    return 2
    # logger.error(evt_timestamp=timestamp, period=[(period.start_timestamp, period.end_timestamp) for period in periods])
    raise DeserializationError(f'Unknown timestamp {timestamp}')


def parse_coordinates(coordinates: list[float], fidelity_version: int) -> Point:
    """Parse coordinates into a kloppy Point.

    Coordinates are cell-based, so 1,1 (low-granularity) or 0.1,0.1
    (high-granularity) is the top-left square 'yard' of the field (in
    landscape), even though 0,0 is the true coordinate of the corner flag.

    [1, 120] x [1, 80]
    +-----+-----+
    | 1,1 | 2,1 |
    +-----+-----+
    | 1,2 | 2,2 |
    +-----+-----+
    """
    cell_side = 0.1 if fidelity_version == 2 else 1.0
    cell_relative_center = cell_side / 2
    if len(coordinates) == 2:
        return Point(
            x=coordinates[0] - cell_relative_center,
            y=coordinates[1] - cell_relative_center,
        )
    elif len(coordinates) == 3:
        # A coordinate in the goal frame, only used for the end location of
        # Shot events. The y-coordinates and z-coordinates are always detailed
        # to a tenth of a yard.
        return Point3D(
            x=coordinates[0] - cell_relative_center,
            y=coordinates[1] - 0.05,
            z=coordinates[2] - 0.05,
        )
    else:
        raise DeserializationError(f'Unknown coordinates format: {coordinates}')



from selenium import webdriver
from selenium.common.exceptions import (
    StaleElementReferenceException,
    TimeoutException,
    WebDriverException,
)
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.webdriver import WebDriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

from ..core.logging import get_logger

logger = get_logger(__name__)


def get_drive_file_list(folder_id: str, retries: int = 3) -> dict[str, str]:
    """Scrapes a Google Drive folder to extract file IDs and names.

    Uses Selenium to load and scroll through all files in a Google Drive folder,
    extracting their IDs and filenames. The function scrolls through the entire
    folder contents and stops when no new files are loaded after multiple attempts.

    Args:
        folder_id: The Google Drive folder ID to scrape.
        retries: The number of retries to attempt if the page does not load.
    Returns:
        A dictionary mapping filenames (without extensions) to file IDs.
    """
    seen_ids = set()
    same_count_repeats = 0
    all_files = {}

    url = f'https://drive.google.com/drive/folders/{folder_id}'

    logger.info('Starting Google Drive folder scrape', folder_id=folder_id, url=url)

    options = Options()
    options.add_argument('--headless=new')
    options.add_argument('--disable-gpu')
    options.add_argument('--window-size=900x6000')

    driver: WebDriver | None = None

    try:
        driver = webdriver.Chrome(options=options)

        driver.get(url)
        logger.info('Navigated to Google Drive folder', url=url)

        while same_count_repeats < retries:
            try:
                # Use explicit wait instead of time.sleep
                WebDriverWait(driver, 10).until(
                    EC.presence_of_element_located(
                        (By.CSS_SELECTOR, 'c-wiz > div:first-of-type[data-id]')
                    )
                )
            except TimeoutException:
                logger.warning(
                    'Timeout waiting for files to load', attempt=same_count_repeats + 1
                )
                same_count_repeats += 1
                continue

            try:
                files = driver.find_elements(
                    By.CSS_SELECTOR, 'c-wiz > div:first-of-type[data-id]'
                )
            except StaleElementReferenceException:
                logger.warning('Stale element reference, retrying')
                continue
            except WebDriverException as e:
                logger.error('WebDriver error while finding elements', error=str(e))
                break

            current_ids = {
                f.get_attribute('data-id') for f in files if f.get_attribute('data-id')
            }

            newly_loaded = current_ids - seen_ids

            for el in files:
                file_id = el.get_attribute('data-id')
                if not file_id:
                    continue

                try:
                    filename = el.find_element(
                        By.CSS_SELECTOR, 'div[data-column-field="6"] > div:last-of-type'
                    ).text
                    all_files[file_id] = filename
                except WebDriverException as e:
                    logger.warning(
                        'Failed to extract filename for file',
                        file_id=file_id,
                        error=str(e),
                    )

            if not newly_loaded:
                same_count_repeats += 1
            else:
                same_count_repeats = 0

            seen_ids.update(current_ids)

            if files:
                try:
                    driver.execute_script(
                        "arguments[0].scrollIntoView({behavior: 'auto', block: 'end'});",
                        files[-1],
                    )
                except WebDriverException as e:
                    logger.error('Failed to scroll', error=str(e))
                    break
    except Exception as e:
        logger.exception('Unexpected error during Google Drive scraping', error=str(e))
        raise
    finally:
        if driver:
            driver.quit()
            logger.debug('WebDriver closed')

    final_file_id_list = {}
    for id_, filename in all_files.items():
        match_id = filename.split('.')[0]
        final_file_id_list[match_id] = id_

    logger.info(
        'Google Drive folder scrape complete',
        folder_id=folder_id,
        total_files=len(final_file_id_list),
    )
    return final_file_id_list



from __future__ import annotations

from datetime import datetime
from enum import Enum

from pydantic import BaseModel, ConfigDict, Field


class PFFEventAdvantageType(str, Enum):
    NON_EVENT = 'N'


class PFFEventBallCarryType(str, Enum):
    CARRY = 'C'
    DRIBBLE = 'D'
    TOUCH = 'T'


class PFFEventBallCarryOutcomeType(str, Enum):
    LEADS_INTO_CHALLENGE = 'C'
    BALL_LOSS = 'L'
    RETAIN = 'R'
    STOPPAGE = 'S'


class PFFEventBetterOptionType(str, Enum):
    BALL_CARRY = 'B'
    CROSS = 'C'
    HOLD = 'H'
    CLEARANCE = 'L'
    CONTINUE = 'O'
    PASS = 'P'
    SHOT = 'S'


class PFFEventBodyMovementType(str, Enum):
    AWAY_FROM_GOAL = 'AG'
    LATERALLY = 'LA'
    STATIC = 'ST'
    TOWARDS_GOAL = 'TG'


class PFFEventBodyType(str, Enum):
    BACK = 'BA'
    BOTTOM = 'BO'
    TWO_HAND_CATCH = 'CA'
    CHEST = 'CH'
    HEAD = 'HE'
    LEFT_FOOT = 'L'
    LEFT_ARM = 'LA'
    LEFT_BACK_HEEL = 'LB'
    LEFT_SHOULDER = 'LC'
    LEFT_HAND = 'LH'
    LEFT_KNEE = 'LK'
    LEFT_SHIN = 'LS'
    LEFT_THIGH = 'LT'
    TWO_HAND_PALM = 'PA'
    TWO_HAND_PUNCH = 'PU'
    RIGHT_FOOT = 'R'
    RIGHT_ARM = 'RA'
    RIGHT_BACK_HEEL = 'RB'
    RIGHT_SHOULDER = 'RC'
    RIGHT_HAND = 'RH'
    RIGHT_KNEE = 'RK'
    RIGHT_SHIN = 'RS'
    RIGHT_THIGH = 'RT'
    TWO_HANDS = 'TWOHANDS'
    VIDEO_MISSING = 'VM'


class PFFEventCarryType(str, Enum):
    LINE_BREAK = 'B'
    CHANGE_OF_DIRECTION = 'C'
    DRIVE_WITH_INTENT = 'D'
    LONG_CARRY = 'L'


class PFFEventChallengeOutcomeType(str, Enum):
    DISTRIBUTION_DISRUPTED = 'B'
    FORCED_OUT_OF_PLAY = 'C'
    DISTRIBUTES_BALL = 'D'  # tem um passe/clearance/shot/cross na hora do challenge
    FOUL = 'F'
    SHIELDS_IN_PLAY = 'I'
    KEEPS_BALL_WITH_CONTACT = 'K'
    ROLLS = 'L'  # girar em volta do marcador (geralmente tem pressao)
    BEATS_MAN_LOSES_BALL = 'M'
    NO_WIN_KEEP_BALL = 'N'  # tentou driblar, nao conseguiu passar e manteve a posse
    OUT_OF_PLAY = 'O'
    PLAYER = 'P'  # a bola vai para um jogador que nao tava envolvido (ex: espirrou)
    RETAIN = 'R'
    SHIELDS_OUT_OF_PLAY = 'S'


class PFFEventChallengeType(str, Enum):
    AERIAL_DUEL = 'A'
    FROM_BEHIND = 'B'
    DRIBBLE = 'D'
    FIFTY = 'FIFTY'
    GK_SMOTHERS = 'G'
    SHIELDING = 'H'
    HAND_TACKLE = 'K'
    SLIDE_TACKLE = 'L'
    SHOULDER_TO_SHOULDER = 'S'
    STANDING_TACKLE = 'T'


class PFFEventClearanceOutcomeType(str, Enum):
    INADVERTENT_SHOT_AT_GOAL = 'A'
    BLOCK = 'B'
    INADVERTENT_SHOT_OWN_GOAL = 'D'
    DELIBERATE_TO_OPPOSITION = 'E'
    OWN_POST = 'N'
    OUT_OF_PLAY = 'O'
    PLAYER = 'P'
    STOPPAGE = 'S'
    DELIBERATE_TO_TOUCH = 'U'
    POST = 'W'


class PFFEventCrossOutcomeType(str, Enum):
    BLOCKED = 'B'
    COMPLETE = 'C'
    DEFENSIVE_INTERCEPTION = 'D'
    LUCKY_SHOT_AT_GOAL = 'I'
    OUT_OF_PLAY = 'O'
    STOPPAGE = 'S'
    UNTOUCHED = 'U'


# TOOD: CONFIRM IF == BELOW
# AWAY_FROM_DEFENDER = 'A'
# CHECKS_MOVEMENT = 'C'
# LEADS_INTO_CHALLENGE = 'E'
# HEAVY = 'H'
# IN_STRIDE = 'I'
# LIGHT = 'L'
# PRECISE = 'P'
# REDIRECTS = 'R'
# STANDARD = 'S'
class PFFEventCrossAccuracyType(str, Enum):
    UNKNOWN_PFF_TYPE_A = 'A'
    UNKNOWN_PFF_TYPE_C = 'C'
    UNKNOWN_PFF_TYPE_E = 'E'
    UNKNOWN_PFF_TYPE_H = 'H'
    UNKNOWN_PFF_TYPE_I = 'I'
    UNKNOWN_PFF_TYPE_L = 'L'
    UNKNOWN_PFF_TYPE_P = 'P'
    UNKNOWN_PFF_TYPE_R = 'R'
    UNKNOWN_PFF_TYPE_S = 'S'


class PFFEventCrossType(str, Enum):
    DRILLED = 'D'
    FLOATED = 'F'
    SWING_IN = 'I'
    SWING_OUT = 'O'
    PLACED = 'P'


class PFFEventCrossZoneType(str, Enum):
    CENTRAL = 'C'
    FAR_POST = 'F'
    NEAR_POST = 'N'
    SIX_YARD_BOX = 'S'


class PFFEventDribbleOutcomeType(str, Enum):
    KEEPS_BALL_WITH_CONTACT = 'B'
    FOUL = 'F'
    MISHIT = 'H'
    KEEPS_BALL = 'K'
    BEATS_MAN_LOSES_BALL = 'L'
    MISSED_FOUL = 'M'
    FORCED_OUT_OF_PLAY = 'O'
    SUCCESSFUL_TACKLE = 'S'


class PFFEventDribbleType(str, Enum):
    BETWEEN_TWO_DEFENDERS = 'B'
    INSIDE = 'I'
    KNOCKS_IN_FRONT = 'K'
    OUTSIDE = 'O'
    TRICK = 'T'


class PFFEventEndType(str, Enum):
    EXTRA_1 = 'F'
    FIRST = 'FIRST'
    GAME = 'G'
    EXTRA_2 = 'S'
    SECOND = 'SECOND'
    Z_TEST_9 = 'Z'


class PFFEventFacingType(str, Enum):
    BACK_TO_GOAL = 'B'
    GOAL = 'G'
    LATERAL = 'L'


class PFFEventFoulOutcomeType(str, Enum):
    NO_FOUL = 'F'
    NO_WARNING = 'N'
    RED_CARD = 'R'
    SECOND_YELLOW = 'S'
    WARNING = 'W'
    YELLOW_CARD = 'Y'


class PFFEventFoulType(str, Enum):
    ADVANTAGE = 'A'
    INFRINGEMENT = 'I'
    MISSED_INFRINGEMENT = 'M'


class PFFEventEventType(str, Enum):
    FIRST_HALF_KICKOFF = 'FIRSTKICKOFF'
    SECOND_HALF_KICKOFF = 'SECONDKICKOFF'
    EXTRA_1_KICKOFF = 'THIRDKICKOFF'
    EXTRA_2_KICKOFF = 'FOURTHKICKOFF'
    GAME_CLOCK_OBSERVATION = 'CLK'
    END_OF_HALF = 'END'
    GROUND = 'G'  # trave!!! travessao ou bandeira de escanteio. O evento eh vazio.
    PLAYER_OFF = 'OFF'
    PLAYER_ON = 'ON'
    POSSESSION = 'OTB'
    BALL_OUT_OF_PLAY = 'OUT'
    PAUSE_OF_GAME_TIME = 'PAU'
    SUB = 'SUB'
    VIDEO = 'VID'


class PFFEventHeightType(str, Enum):
    ABOVE_HEAD = 'A'
    GROUND = 'G'
    BETWEEN_WAIST_AND_HEAD = 'H'
    OFF_GROUND_BELOW_WAIST = 'L'
    VIDEO_MISSING = 'M'
    HALF_VOLLEY = 'V'  # Bola quica e ta subindo


class PFFEventIncompletionReasonType(str, Enum):
    BEHIND = 'BH'
    BLOCKED = 'BL'
    CAUGHT = 'CA'
    DEFENSIVE_CONTACT = 'CO'
    DELIBERATE = 'DB'
    DEFENSIVE_CHALLENGE = 'DC'
    DEFLECTED = 'DF'
    DEFENDER_INTERCEPTION = 'DI'
    FOUL = 'FO'
    HIGH = 'HI'
    HIT_OFFICIAL = 'HO'
    IN_FRONT = 'IF'
    RECEIVER_LETS_BALL_RUN = 'LB'
    MISCOMMUNICATION = 'MC'
    MISS_HIT = 'MH'
    PASSER_SLIPPED = 'PS'
    RECEIVER_DIDNT_RETURN_TO_BALL = 'RB'
    RECEIVER_SLIPPED = 'RF'
    RECEIVER_MISSES_BALL = 'RM'
    RECEIVER_STOPPED = 'RS'
    REFEREE_IN_WAY = 'RW'
    SPECULATIVE = 'SP'
    UNDERHIT = 'UH'
    VIDEO_MISSING = 'VM'


class PFFEventInitialTouchType(str, Enum):
    H2C_BAD = 'B'
    H2C_GOOD = 'G'
    MISCONTROL = 'M'
    PLUS = 'P'
    STANDARD = 'S'


class PFFEventLinesBrokenType(str, Enum):
    ATT = 'A'
    ATT_DEF = 'AD'
    ATT_MID = 'AM'
    ATT_MID_DEF = 'AMD'
    DEF = 'D'
    MID = 'M'
    MID_DEF = 'MD'


class PFFEventMissedTouchType(str, Enum):
    MISSED_CROSS = 'C'
    DUMMY = 'D'  # deixa a bola passar pra outro
    MISSED_INTERCEPTION = 'I'
    MISSED_CLEARANCE = 'L'
    MISSED_TOUCH = 'M'
    MISSED_SHOT = 'O'
    MISSED_PASS = 'P'
    SLIP = 'S'


class PFFEventOpportunityType(str, Enum):
    CHANCE_CREATED = 'C'
    DANGEROUS_POSITION = 'D'
    HALF_CHANCE = 'H'
    SPACE_TO_CLEAR = 'L'
    NEGATIVE_CHANCE_CREATED = 'N'
    NEGATIVE_DANGEROUS_POSITION = 'P'
    SPACE_TO_CROSS = 'R'
    SPACE_TO_SHOOT = 'S'


class PFFEventOriginateType(str, Enum):
    CORNER_FLAG = 'C'
    MISCELLANEOUS = 'M'
    PLAYER = 'P'
    POST = 'W'


class PFFEventOutType(str, Enum):
    AWAY_SCORE = 'A'
    HOME_SCORE = 'H'
    TOUCH = 'T'
    WHISTLE = 'W'


class PFFEventPassAccuracyType(str, Enum):
    AWAY_FROM_DEFENDER = 'A'
    CHECKS_MOVEMENT = 'C'
    LEADS_INTO_CHALLENGE = 'E'
    HEAVY = 'H'
    IN_STRIDE = 'I'
    LIGHT = 'L'
    PRECISE = 'P'
    REDIRECTS = 'R'
    STANDARD = 'S'


class PFFEventPassOutcomeType(str, Enum):
    BLOCKED = 'B'
    COMPLETE = 'C'
    DEFENSIVE_INTERCEPTION = 'D'
    LUCKY_SHOT_OWN_GOAL = 'G'
    LUCKY_SHOT_GOAL = 'I'
    OUT_OF_PLAY = 'O'
    STOPPAGE = 'S'


class PFFEventPassType(str, Enum):
    CUTBACK = 'B'
    CREATE_CONTEST = 'C'
    FLICK_ON = 'F'
    LONG_THROW = 'H'
    LONG_PASS = 'L'
    MISS_HIT = 'M'
    BALL_OVER_THE_TOP = 'O'
    STANDARD_PASS = 'S'
    THROUGH_BALL = 'T'
    SWITCH = 'W'


class PFFEventPlayerOffType(str, Enum):
    EQUIPMENT = 'E'
    INJURY = 'I'
    MISCELLANEOUS = 'M'
    RED_CARD = 'R'
    # YELLOW_CARD = 'Y'


class PFFEventPositionGroupType(str, Enum):
    ATTACK_MID = 'AM'
    CENTER_FORWARD = 'CF'
    CENTER_MID = 'CM'
    DEFENDER = 'D'
    DEFENSIVE_MID = 'DM'
    FORWARD = 'F'
    GK = 'GK'
    LEFT_BACK = 'LB'
    LEFT_CENTER_BACK = 'LCB'
    LEFT_MID = 'LM'
    LEFT_WINGER = 'LW'
    LEFT_WING_BACK = 'LWB'
    MIDFIELDER = 'M'
    MID_CENTER_BACK = 'MCB'
    RIGHT_BACK = 'RB'
    RIGHT_CENTER_BACK = 'RCB'
    CENTER_BACK = 'CB'
    RIGHT_MID = 'RM'
    RIGHT_WINGER = 'RW'
    RIGHT_WING_BACK = 'RWB'


class PFFEventPossessionEventType(str, Enum):
    BALL_CARRY = 'BC'
    CHALLENGE = 'CH'
    CLEARANCE = 'CL'
    CROSS = 'CR'
    PASS = 'PA'
    REBOUND = 'RE'
    SHOT = 'SH'


class PFFEventPotentialOffenseType(str, Enum):
    DISSENT = 'D'
    OFF_THE_BALL = 'F'
    HAND_BALL = 'H'
    ON_THE_BALL = 'N'
    OFFSIDE = 'O'
    TECHNICAL = 'T'
    DIVA = 'V'


class PFFEventPressureType(str, Enum):
    ATTEMPTED = 'A'
    PASSING_LANE = 'L'
    PRESSURED = 'P'


# Rebound: a bola bateu no jogador
class PFFEventReboundOutcomeType(str, Enum):
    LUCKY_SHOT_GOAL = 'A'
    LUCKY_SHOT_OWN_GOAL = 'D'
    PLAYER = 'P'
    RETAIN = 'R'
    OUT_OF_TOUCH = 'T'


class PFFEventSaveReboundType(str, Enum):
    CROSSBAR = 'CB'
    LEFT_BEHIND_GOAL = 'GL'
    RIGHT_BEHIND_GOAL = 'GR'
    LEFT_BEHIND_GOAL_HIGH = 'HL'
    RIGHT_BEHIND_GOAL_HIGH = 'HR'
    LEFT_SIX_YARD_BOX = 'L6'
    LEFT_AREA = 'LA'
    LEFT_OUT_OF_BOX = 'LO'
    LEFT_POST = 'LP'
    MIDDLE_SIX_YARD_BOX = 'M6'
    MIDDLE_AREA = 'MA'
    MIDDLE_OUT_OF_BOX = 'MO'
    CROSSBAR_OVER = 'OC'
    RIGHT_SIX_YARD_BOX = 'R6'
    RIGHT_AREA = 'RA'
    RIGHT_OUT_OF_BOX = 'RO'
    RIGHT_POST = 'RP'


class PFFEventSetpieceType(str, Enum):
    CORNER = 'C'
    DROP_BALL = 'D'
    FREE_KICK = 'F'
    GOAL_KICK = 'G'
    KICKOFF = 'K'
    PENALTY = 'P'
    THROW_IN = 'T'


class PFFEventShotHeightType(str, Enum):
    BOTTOM_THIRD = 'BOTTOMTHIRD'
    CROSSBAR = 'C'
    SHORT = 'F'
    GROUND = 'G'
    MIDDLE_THIRD = 'MIDDLETHIRD'
    CROSSBAR_NARROW_OVER = 'N'
    OVER = 'O'
    TOP_THIRD = 'TOPTHIRD'
    CROSSBAR_NARROW_UNDER = 'U'
    UNKNOWN_TYPE_PFF_1 = '1'
    UNKNOWN_TYPE_PFF_2 = '2'
    UNKNOWN_TYPE_PFF_3 = '3'


class PFFEventShotNatureType(str, Enum):
    PLACEMENT = 'A'
    FLICK = 'F'
    LACES = 'L'
    POWER = 'P'
    SCUFF = 'S'
    TOE_PUNT = 'T'


class PFFEventShotOutcomeType(str, Enum):
    ON_TARGET_BLOCK = 'B'
    OFF_TARGET_BLOCK = 'C'
    SAVE_OFF_TARGET = 'F'
    GOAL = 'G'
    GOALLINE_CLEARANCE = 'L'
    OFF_TARGET = 'O'
    ON_TARGET = 'S'


class PFFEventShotType(str, Enum):
    BICYCLE = 'B'
    DIVING = 'D'
    SIDE_FOOT = 'F'
    SLIDING = 'I'
    LOB = 'L'
    OUTSIDE_FOOT = 'O'
    STANDARD = 'S'
    STUDS = 'T'
    VOLLEY = 'V'


class PFFEventStadiumGrassType(str, Enum):
    ASTRO_TURF = 'A'
    FIELD_TURF = 'F'
    REAL = 'R'
    NATURAL = 'N'


class PFFEventStadiumType(str, Enum):
    CONVERSION = 'C'
    DOMED = 'D'
    INDOOR = 'I'
    OUTDOOR = 'O'


class PFFEventSubType(str, Enum):
    BLOOD = 'B'
    SIN_BIN_COVER = 'C'
    HEAD_INJURY_ASSESSMENT = 'H'
    RETURN_FROM_HEAD_INJURY_ASSESSMENT = 'R'
    STANDARD = 'S'


class PFFEventTackleAttemptType(str, Enum):
    DELIBERATE_FOUL = 'D'
    NO_TACKLE_FAKE_EVENT = 'F'
    GO_FOR_BALL = 'G'
    NO_TACKLE = 'T'


class PFFEventTouchOutcomeType(str, Enum):
    INADVERTENT_SHOT_AT_GOAL = 'A'
    CHALLENGE = 'C'
    INADVERTENT_SHOT_OWN_GOAL = 'D'
    GOAL = 'G'
    OUT_OF_PLAY = 'O'
    PLAYER = 'P'
    RETAIN = 'R'
    OWN_GOAL = 'W'


class PFFEventTouchType(str, Enum):
    BALL_IN_HAND = 'B'
    FAILED_CROSS = 'C'
    HAND_BALL = 'D'
    FAILED_TRAP = 'F'
    FAILED_CATCH = 'G'
    HEAVY_TOUCH = 'H'
    FAILED_CLEARANCE = 'L'
    FAILED_PASS = 'P'
    FAILED_SHOT = 'S'
    TAKE_OVER = 'T'


class PFFEventVarReasonType(str, Enum):
    MISSED = 'I'
    OVERTURN = 'O'


class PFFEventVideoAngleType(str, Enum):
    BAD_ANGLE = 'B'
    MISSING = 'M'


class PFFEventId(BaseModel):
    model_config = ConfigDict(extra='forbid')
    id: str


class PFFEventPlayerIdNickname(BaseModel):
    model_config = ConfigDict(extra='forbid')

    id: str
    # nickname: str
    # team: PFFEventId


class PFFEventBallCarryEvent(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    game: PFFEventId
    gameEvent: PFFEventId
    possessionEvent: PFFEventId

    badParry: bool
    ballCarrierPlayer: PFFEventPlayerIdNickname | None
    ballCarryOutcome: PFFEventBallCarryOutcomeType | None
    ballCarryType: PFFEventBallCarryType | None
    betterOptionPlayer: PFFEventPlayerIdNickname | None
    betterOptionTime: str | None
    betterOptionType: PFFEventBetterOptionType | None
    carryIntent: str | None
    carryType: PFFEventCarryType | None
    challengerPlayer: PFFEventPlayerIdNickname | None
    clearerPlayer: PFFEventPlayerIdNickname | None
    createsSpace: bool | None
    defenderPlayer: PFFEventPlayerIdNickname | None
    dribbleOutcomeType: PFFEventDribbleOutcomeType | None
    keeperPlayer: PFFEventPlayerIdNickname | None
    linesBrokenType: PFFEventLinesBrokenType | None
    missedTouchPlayer: PFFEventPlayerIdNickname | None
    missedTouchType: PFFEventMissedTouchType | None
    opportunityType: PFFEventOpportunityType | None
    period: int | None
    pressurePlayer: PFFEventPlayerIdNickname | None
    pressureType: PFFEventPressureType
    saveReboundType: PFFEventSaveReboundType | None
    saveable: bool
    shotInitialHeightType: PFFEventShotHeightType | None
    shotOutcomeType: PFFEventShotOutcomeType | None
    successful: bool | None
    tackleAttemptType: PFFEventTackleAttemptType | None
    touchOutcomeType: PFFEventTouchOutcomeType | None
    touchType: PFFEventTouchType | None


class PFFEventCacheStats(BaseModel):
    model_config = ConfigDict(extra='forbid')
    hitRate: float
    name: str


class PFFEventChallengeEvent(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    game: PFFEventId
    gameEvent: PFFEventId
    possessionEvent: PFFEventId

    additionalChallenger1: PFFEventPlayerIdNickname | None
    additionalChallenger2: PFFEventPlayerIdNickname | None
    additionalChallenger3: PFFEventPlayerIdNickname | None
    advantageType: PFFEventAdvantageType | None
    ballCarrierPlayer: PFFEventPlayerIdNickname | None
    challengeOutcomeType: PFFEventChallengeOutcomeType | None
    challengeType: PFFEventChallengeType | None
    challengeWinnerPlayer: PFFEventPlayerIdNickname | None
    challengerAwayPlayer: PFFEventPlayerIdNickname | None
    challengerHomePlayer: PFFEventPlayerIdNickname | None
    challengerPlayer: PFFEventPlayerIdNickname | None
    createsSpace: bool | None
    dribbleType: PFFEventDribbleType | None
    keeperPlayer: PFFEventPlayerIdNickname | None
    linesBrokenType: PFFEventLinesBrokenType | None
    missedTouchPlayer: PFFEventPlayerIdNickname | None
    missedTouchType: PFFEventMissedTouchType | None
    opportunityType: PFFEventOpportunityType | None
    period: str | None
    pressurePlayer: PFFEventPlayerIdNickname | None
    tackleAttemptType: PFFEventTackleAttemptType | None
    trickType: str | None


class PFFEventClearanceEvent(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    game: PFFEventId
    gameEvent: PFFEventId
    possessionEvent: PFFEventId

    advantageType: PFFEventAdvantageType | None
    badParry: bool
    ballHeightType: PFFEventHeightType | None
    betterOptionPlayer: PFFEventPlayerIdNickname | None
    betterOptionTime: str | None
    betterOptionType: PFFEventBetterOptionType | None
    blockerPlayer: PFFEventPlayerIdNickname | None
    clearanceBodyType: PFFEventBodyType | None
    clearanceOutcomeType: PFFEventClearanceOutcomeType
    clearancePlayer: PFFEventPlayerIdNickname
    createsSpace: bool | None
    createsSpaceVsPlayer: PFFEventPlayerIdNickname | None
    failedInterventionPlayer: PFFEventPlayerIdNickname | None
    failedInterventionPlayer1: PFFEventPlayerIdNickname | None
    failedInterventionPlayer2: PFFEventPlayerIdNickname | None
    failedInterventionPlayer3: PFFEventPlayerIdNickname | None
    heightType: PFFEventHeightType | None
    keeperPlayer: PFFEventPlayerIdNickname | None
    missedTouchPlayer: PFFEventPlayerIdNickname | None
    missedTouchType: PFFEventMissedTouchType | None
    opportunityType: PFFEventOpportunityType | None
    period: str | None
    pressurePlayer: PFFEventPlayerIdNickname | None
    pressureType: PFFEventPressureType | None
    saveReboundType: PFFEventSaveReboundType | None
    saveable: bool
    shotInitialHeightType: PFFEventShotHeightType | None
    shotOutcomeType: PFFEventShotOutcomeType | None


class PFFEventCompetition(BaseModel):
    model_config = ConfigDict(extra='forbid')
    availableSeasons: list[PFFEventCompetitionSeason] = Field(default_factory=list)
    id: str
    games: list[PFFEventGame] = Field(default_factory=list)
    name: str
    seasonGames: list[PFFEventGame] = Field(default_factory=list)
    teams: list[PFFEventTeam] = Field(default_factory=list)


class PFFEventCompetitionSeason(BaseModel):
    model_config = ConfigDict(extra='forbid')
    season: str
    start: str
    end: str


class PFFEventConfederation(BaseModel):
    model_config = ConfigDict(extra='forbid')
    id: str
    abbreviation: str
    name: str


class PFFEventCrossEvent(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    game: PFFEventId
    gameEvent: PFFEventId
    possessionEvent: PFFEventId

    advantageType: PFFEventAdvantageType | None
    badParry: bool
    ballHeightType: PFFEventHeightType
    betterOptionPlayer: PFFEventPlayerIdNickname | None
    betterOptionTime: str | None
    betterOptionType: PFFEventBetterOptionType | None
    blockerPlayer: PFFEventPlayerIdNickname | None
    clearerPlayer: PFFEventPlayerIdNickname | None
    completeToPlayer: PFFEventPlayerIdNickname | None
    createsSpace: bool | None
    createsSpaceVsPlayer: PFFEventPlayerIdNickname | None
    crossAccuracyType: PFFEventCrossAccuracyType | None
    crossHighPointType: PFFEventHeightType | None
    crossOutcomeType: PFFEventCrossOutcomeType | None
    crossType: PFFEventCrossType | None
    crossZoneType: PFFEventCrossZoneType | None
    crosserBodyType: PFFEventBodyType
    crosserPlayer: PFFEventPlayerIdNickname
    defenderBallHeightType: PFFEventHeightType | None
    defenderBodyType: PFFEventBodyType | None
    defenderPlayer: PFFEventPlayerIdNickname | None
    deflectorBodyType: PFFEventBodyType | None
    deflectorPlayer: PFFEventPlayerIdNickname | None
    failedInterventionPlayer: PFFEventPlayerIdNickname | None
    failedInterventionPlayer1: PFFEventPlayerIdNickname | None
    failedInterventionPlayer2: PFFEventPlayerIdNickname | None
    failedInterventionPlayer3: PFFEventPlayerIdNickname | None
    incompletionReasonType: PFFEventIncompletionReasonType | None
    intendedTargetPlayer: PFFEventPlayerIdNickname | None
    keeperPlayer: PFFEventPlayerIdNickname | None
    late: bool | None
    missedTouchPlayer: PFFEventPlayerIdNickname | None
    missedTouchType: PFFEventMissedTouchType | None
    noLook: bool
    opportunityType: PFFEventOpportunityType | None
    period: str | None
    pressurePlayer: PFFEventPlayerIdNickname | None
    pressureType: PFFEventPressureType | None
    receiverBallHeightType: PFFEventHeightType | None
    receiverBodyType: PFFEventBodyType | None
    saveReboundType: PFFEventSaveReboundType | None
    saveable: bool
    secondIncompletionReasonType: PFFEventIncompletionReasonType | None
    shotInitialHeightType: PFFEventShotHeightType | None
    shotOutcomeType: PFFEventShotOutcomeType | None


class PFFEventDefender(BaseModel):
    model_config = ConfigDict(extra='forbid')

    id: str

    defenderPlayer: PFFEventPlayerIdNickname
    game: PFFEventId
    gameEvent: PFFEventId
    possessionEvent: PFFEventId


class PFFEventFederation(BaseModel):
    model_config = ConfigDict(extra='forbid')

    id: str
    name: str
    englishName: str
    abbreviation: str
    confederation: PFFEventConfederation
    country: str


class PFFEventFoul(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    game: PFFEventId
    gameEvent: PFFEventId
    possessionEvent: PFFEventId

    badCall: bool | None
    correctDecision: bool
    culpritPlayer: PFFEventPlayerIdNickname | None
    foulOutcomeType: PFFEventFoulOutcomeType | None
    foulType: PFFEventFoulType | None
    potentialOffenseType: PFFEventPotentialOffenseType | None
    sequence: int | None
    tacticalFoul: bool | None
    var: bool | None
    varCulpritPlayer: PFFEventPlayerIdNickname | None
    varOutcomeType: PFFEventFoulOutcomeType | None
    varPotentialOffenseType: PFFEventPotentialOffenseType | None
    varReasonType: PFFEventVarReasonType | None
    victimPlayer: PFFEventPlayerIdNickname | None


class PFFEventLocation(BaseModel):
    model_config = ConfigDict(extra='ignore')
    id: str

    eventModule: str
    # ballCarryEvent: PFFEventBallCarryEvent
    # challengeEvent: PFFEventChallengeEvent
    # clearanceEvent: PFFEventClearanceEvent
    # crossEvent: PFFEventCrossEvent
    # gameEvent: PFFEventId # Deveri incluir...
    name: str
    # passingEvent: PFFEventPassingEvent
    # possessionEvent: PFFEventId # Deveria incruir....
    # reboundEvent: PFFEventReboundEvent
    # shootingEvent: PFFEventShootingEvent
    x: float
    y: float


class PFFEventNation(BaseModel):
    model_config = ConfigDict(extra='forbid')
    country: str
    federation: PFFEventFederation
    fifaCode: str
    id: str
    iocCode: str


class PFFEventPassingEvent(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    game: PFFEventId
    gameEvent: PFFEventId
    possessionEvent: PFFEventId

    advantageType: PFFEventAdvantageType | None
    badParry: bool
    ballHeightType: PFFEventHeightType | None
    betterOptionPlayer: PFFEventPlayerIdNickname | None
    betterOptionTime: str | None
    betterOptionType: PFFEventBetterOptionType | None
    blockerPlayer: PFFEventPlayerIdNickname | None
    clearerPlayer: PFFEventPlayerIdNickname | None
    createsSpace: bool | None
    createsSpaceVsPlayer: PFFEventPlayerIdNickname | None
    defenderBodyType: PFFEventBodyType | None
    defenderHeightType: PFFEventHeightType | None
    defenderPlayer: PFFEventPlayerIdNickname | None
    deflectorBodyType: PFFEventBodyType | None
    deflectorPlayer: PFFEventPlayerIdNickname | None
    failedInterventionPlayer: PFFEventPlayerIdNickname | None
    failedInterventionPlayer1: PFFEventPlayerIdNickname | None
    failedInterventionPlayer2: PFFEventPlayerIdNickname | None
    failedInterventionPlayer3: PFFEventPlayerIdNickname | None
    incompletionReasonType: PFFEventIncompletionReasonType | None
    keeperPlayer: PFFEventPlayerIdNickname | None
    late: bool | None
    linesBrokenType: PFFEventLinesBrokenType | None
    missedTouchPlayer: PFFEventPlayerIdNickname | None
    missedTouchType: PFFEventMissedTouchType | None
    noLook: bool | None
    opportunityType: PFFEventOpportunityType | None
    passAccuracyType: PFFEventPassAccuracyType | None
    passBodyType: PFFEventBodyType | None
    passHighPointType: PFFEventHeightType | None
    passOutcomeType: PFFEventPassOutcomeType | None
    passType: PFFEventPassType | None
    passerPlayer: PFFEventPlayerIdNickname | None
    period: str | None
    pressurePlayer: PFFEventPlayerIdNickname | None
    pressureType: PFFEventPressureType | None
    receiverBodyType: PFFEventBodyType | None
    receiverFacingType: PFFEventFacingType | None
    receiverHeightType: PFFEventHeightType | None
    receiverPlayer: PFFEventPlayerIdNickname | None
    saveReboundType: PFFEventSaveReboundType | None
    saveable: bool
    secondIncompletionReasonType: PFFEventIncompletionReasonType | None
    shotInitialHeightType: PFFEventShotHeightType | None
    shotOutcomeType: PFFEventShotOutcomeType | None
    targetFacingType: PFFEventFacingType
    targetPlayer: PFFEventPlayerIdNickname | None


class PFFEventPlayer(BaseModel):
    model_config = ConfigDict(extra='forbid')
    id: str
    dob: str | None
    firstName: str
    lastName: str
    nickname: str
    height: float | None  # in cm
    weight: float | None
    gender: str | None
    nationality: PFFEventNation | None
    positionGroupType: PFFEventPositionGroupType
    preferredFoot: str | None
    secondNationality: PFFEventNation | None
    countryOfBirth: PFFEventNation | None


class PFFEventPossessionEvent(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    game: PFFEventId
    gameEvent: PFFEventId

    ballCarryEvent: PFFEventBallCarryEvent | None
    challengeEvent: PFFEventChallengeEvent | None
    clearanceEvent: PFFEventClearanceEvent | None
    crossEvent: PFFEventCrossEvent | None
    reboundEvent: PFFEventReboundEvent | None
    shootingEvent: PFFEventShootingEvent | None
    passingEvent: PFFEventPassingEvent | None

    defenders: list[PFFEventDefender] = Field(default_factory=list)
    endTime: float | None
    formattedGameClock: str | None
    fouls: list[PFFEventFoul] = Field(default_factory=list)
    gameClock: float | None
    lastInGameEvent: int
    period: str | None
    possessionEventType: PFFEventPossessionEventType
    startTime: float
    videoUrl: str


class PFFEventReboundEvent(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    game: PFFEventId
    gameEvent: PFFEventId
    possessionEvent: PFFEventId

    advantageType: PFFEventAdvantageType | None
    badParry: bool
    blockerPlayer: PFFEventPlayerIdNickname | None
    # clearerPlayer: PFFEventPlayerIdNickname | None
    keeperPlayer: PFFEventPlayerIdNickname | None
    missedTouchPlayer: PFFEventPlayerIdNickname | None
    missedTouchType: PFFEventMissedTouchType | None
    originateType: PFFEventOriginateType
    period: str | None
    reboundBodyType: PFFEventBodyType | None
    reboundHeightType: PFFEventHeightType | None
    reboundHighPointType: PFFEventHeightType | None
    reboundOutcomeType: PFFEventReboundOutcomeType | None
    rebounderPlayer: PFFEventPlayerIdNickname | None
    saveReboundType: PFFEventSaveReboundType | None
    saveable: bool
    shotInitialHeightType: PFFEventShotHeightType | None
    shotOutcomeType: PFFEventShotOutcomeType | None


class PFFEventRoster(BaseModel):
    model_config = ConfigDict(extra='forbid')

    id: str
    game: PFFEventId
    player: PFFEventPlayer
    positionGroupType: PFFEventPositionGroupType
    shirtNumber: int
    started: bool
    team: PFFEventTeam


class PFFEventShootingEvent(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    game: PFFEventId
    gameEvent: PFFEventId
    possessionEvent: PFFEventId

    advantageType: PFFEventAdvantageType | None
    badParry: bool
    ballHeightType: PFFEventHeightType | None
    ballMoving: bool
    betterOptionPlayer: PFFEventPlayerIdNickname | None
    betterOptionTime: str | None
    betterOptionType: PFFEventBetterOptionType | None
    blockerPlayer: PFFEventPlayerIdNickname | None
    bodyMovementType: PFFEventBodyMovementType | None
    clearerPlayer: PFFEventPlayerIdNickname | None
    createsSpace: bool | None
    createsSpaceVsPlayer: PFFEventPlayerIdNickname | None
    deflectorBodyType: PFFEventBodyType | None
    deflectorPlayer: PFFEventPlayerIdNickname | None
    failedInterventionPlayer: PFFEventPlayerIdNickname | None
    failedInterventionPlayer1: PFFEventPlayerIdNickname | None
    failedInterventionPlayer2: PFFEventPlayerIdNickname | None
    failedInterventionPlayer3: PFFEventPlayerIdNickname | None
    keeperTouchType: PFFEventBodyType | None
    late: bool | None
    missedTouchPlayer: PFFEventPlayerIdNickname | None
    missedTouchType: PFFEventMissedTouchType | None
    noLook: bool
    period: str | None
    pressurePlayer: PFFEventPlayerIdNickname | None
    pressureType: PFFEventPressureType | None
    saveHeightType: PFFEventShotHeightType | None
    saveReboundType: PFFEventSaveReboundType | None
    saveable: bool | None
    saverPlayer: PFFEventPlayerIdNickname | None
    shooterPlayer: PFFEventPlayerIdNickname | None
    shotBodyType: PFFEventBodyType | None
    shotInitialHeightType: PFFEventShotHeightType | None
    shotNatureType: PFFEventShotNatureType | None
    shotOutcomeType: PFFEventShotOutcomeType | None
    shotType: PFFEventShotType


class PFFEventPitch(BaseModel):
    model_config = ConfigDict(extra='forbid')

    length: float
    width: float
    startDate: datetime
    endDate: datetime
    # grassType: PFFEventStadiumGrassType


class PFFEventStadium(BaseModel):
    model_config = ConfigDict(extra='forbid')
    id: str
    name: str
    pitches: list[PFFEventPitch] = Field(default_factory=list)


class PFFEventTeam(BaseModel):
    model_config = ConfigDict(extra='forbid')
    id: str
    name: str
    shortName: str


class PFFEventKit(BaseModel):
    model_config = ConfigDict(extra='forbid')
    id: str
    name: str
    primaryColor: str
    primaryTextColor: str
    secondaryColor: str
    secondaryTextColor: str


class PFFEventVideo(BaseModel):
    duration: float
    fps: float
    # game: PFFEventId
    id: str
    originalFilename: str
    # status: int
    videoUrl: str


class PFFEventGameEvent(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str

    advantageType: PFFEventAdvantageType | None
    bodyType: PFFEventBodyType | None
    defenderLocations: list[PFFEventLocation] | None = Field(default_factory=list)
    duration: float
    earlyDistribution: bool | None
    endTime: float | None
    endType: PFFEventEndType | None
    facingType: PFFEventFacingType | None
    formattedGameClock: str
    game: PFFEventId
    gameClock: float
    gameEventType: PFFEventEventType
    heightType: PFFEventHeightType
    initialTouchType: PFFEventInitialTouchType | None
    offenderLocations: list[PFFEventLocation] | None = Field(default_factory=list)
    otherPlayer: PFFEventPlayerIdNickname | None
    outType: PFFEventOutType | None
    period: int | None
    player: PFFEventPlayerIdNickname | None
    playerOff: PFFEventPlayerIdNickname | None
    playerOffType: PFFEventPlayerOffType | None
    playerOn: PFFEventPlayerIdNickname | None
    possessionEvents: list[PFFEventPossessionEvent] = Field(default_factory=list)
    pressurePlayer: PFFEventPlayerIdNickname | None
    pressureType: PFFEventPressureType | None
    setpieceType: PFFEventSetpieceType | None
    startTime: float
    subType: PFFEventSubType
    team: PFFEventTeam | None
    touches: int | None
    touchesInBox: int | None
    # video: PFFEventVideo | None
    videoAngleType: PFFEventVideoAngleType | None
    # videoMissing: bool | None
    videoUrl: str


class PFFEventGame(BaseModel):
    model_config = ConfigDict(extra='forbid')
    id: str
    awayTeam: PFFEventTeam
    homeTeam: PFFEventTeam
    awayTeamKit: PFFEventKit
    homeTeamKit: PFFEventKit
    competition: PFFEventCompetition
    complete: bool
    date: datetime
    endPeriod1: float
    endPeriod2: float
    gameEvents: list[PFFEventGameEvent] = Field(default_factory=list)
    halfPeriod: float
    homeTeamStartLeft: bool
    homeTeamStartLeftExtraTime: bool
    period1: float
    period2: float
    rosters: list[PFFEventRoster] = Field(default_factory=list)
    season: str | int
    stadium: PFFEventStadium
    startPeriod1: float
    startPeriod2: float
    venue: str | None
    videos: list[PFFEventVideo] = Field(default_factory=list)
    week: int



"""Implements functionality for watching videos of possession events."""

from typing import cast
from uuid import uuid4

from bs4 import Tag
from IPython import get_ipython  #  type: ignore
from IPython.display import HTML, display

from ..core.config import get_config_value
from ..html_templates import env


def _scrape_playlist_from_pff_video(url: str) -> str:
    """Scrape PFF video details from a given URL.

    Parameters:
    -----------
    url : HttpUrl
        The URL we want to scrape.

    Returns:
    --------
    str
        The url to the m3u8 playlist.
    """
    import requests
    from bs4 import BeautifulSoup

    response = requests.get(url)

    if response.status_code != 200:
        raise ValueError(f'Failed to load video. status_code={response.status_code}')

    soup = BeautifulSoup(response.text, 'html.parser')
    video = soup.find('video')

    if not video:
        raise ValueError('Failed to find the <video> element in the page.')

    video = cast(Tag, video)

    return video.get('data-playlist-url')  # type: ignore


def video(
    event: dict,
    *,
    width: int | str = '80%',
    height: int | str = 'auto',
    start_buffer_sec: float | None = None,
    end_buffer_sec: float | None = None,
) -> None:
    """Display a video for a PFF_PossessionEvent on IPython environments."""
    buffer = get_config_value('visualization', 'video_buffer_sec')
    if start_buffer_sec is None:
        start_buffer_sec = buffer
    if end_buffer_sec is None:
        end_buffer_sec = buffer

    if not get_ipython():
        raise RuntimeError('This function is only available on IPython environments')

    if not event['gameEvents']['videoUrl']:
        raise ValueError('No video URL found for this event.')

    # this is temporary fix as PFF is serving the url without the film_room
    split_url = event['gameEvents']['videoUrl'].split('/')
    event_path = '/'.join(split_url[1:])
    # film_room_url = f'https://epitome-staging.pff.com/en/film_room/{event_path}'
    film_room_url = event['gameEvents']['videoUrl']

    playlist_url = _scrape_playlist_from_pff_video(film_room_url)
    blob_url = f'blob:https://epitome-staging.pff.com/{split_url[1]}'

    video_vars = {
        'width': width,
        'height': height,
        'event_id': f'{event['gameEventId']}-{uuid4()}',
        'start_position': float(event['startTime']) - start_buffer_sec,  # type: ignore
        'end_position': float(event['endTime']) + end_buffer_sec,  # type: ignore
        'playlist_url': playlist_url,
        'video_url': blob_url,
    }

    video_template = env.get_template('video.jinja2')

    display(HTML(video_template.render(video_vars)))



"""Transform raw PFF event data into flattened structures for DataFrames.

This module provides functionality to transform PFF event data into flat structures
suitable for analysis in pandas DataFrames. It handles both game-level events and
nested event data like possessions, ball carries, challenges, etc.
"""

from enum import Enum
from typing import Any

from ..core.logging import get_logger
from .pff_events_mappings import EVENTS_FIELD_MAP

logger = get_logger(__name__)


POSSESSION_EVENT_TYPES = [
    'ballCarryEvent',
    'challengeEvent',
    'clearanceEvent',
    'crossEvent',
    'reboundEvent',
    'shootingEvent',
    'passingEvent',
]


class TransformFormat(str, Enum):
    """Format options for the transformed output."""

    WIDE = 'wide'  # One row per event with nested data as columns
    LONG = 'long'  # Multiple rows per event (one per possession/sub-event)


def _flatten_dict(d: dict[str, Any], prefix: str = '') -> dict[str, Any]:
    """
    Recursively flatten a nested dictionary with prefixed keys.

    Args:
        d: Dictionary to flatten
        prefix: Prefix to add to keys (with trailing underscore if non-empty)

    Returns:
        Flattened dictionary where nested keys are combined with their parent keys
    """
    result = {}

    try:
        for k, v in d.items():
            key = f'{prefix}_{k}' if prefix else k

            if isinstance(v, dict):
                result.update(_flatten_dict(v, key))
            elif isinstance(v, list):
                # For lists, add a count and flatten each item with an index
                result[f'{key}_count'] = len(v)
                for i, item in enumerate(v):
                    if isinstance(item, dict):
                        result.update(_flatten_dict(item, f'{key}_{i}'))
                    else:
                        result[f'{key}_{i}'] = item
            else:
                result[key] = v
    except AttributeError:
        logger.error(f'Error flattening dictionary: {d}')
        raise

    return result


def transform_match_metadata(match: dict[str, Any]) -> dict[str, Any]:
    """Extract match metadata into a flat dictionary."""
    metadata_field_map = {
        # Match identification
        'id': 'match_id',
        'date': 'match_date',
        'week': 'match_week',
        'venue': 'match_venue',
        'season': 'season',
        # Competition
        'competition_id': 'competition_id',
        'competition_name': 'competition_name',
        # Stadium
        'stadium_id': 'stadium_id',
        'stadium_name': 'stadium_name',
        'stadium_pitch_length': 'pitch_length',
        'stadium_pitch_width': 'pitch_width',
        'stadium_pitch_grass_type': 'pitch_grass_type',
        # Video
        'video_fps': 'video_fps',
        # Match timing
        'startPeriod1': 'period_1_start',
        'endPeriod1': 'period_1_end',
        'startPeriod2': 'period_2_start',
        'endPeriod2': 'period_2_end',
        # Helps with conversions and maybe tracking
        'homeTeamStartLeft': 'home_start_left',
        'homeTeamStartLeftExtraTime': 'home_start_left_extra',
        'homeTeam_id': 'home_team_id',
        'homeTeam_name': 'home_team_name',
        'homeTeamKit_primaryColor': 'home_kit_primary_color',
        'homeTeamKit_primaryTextColor': 'home_kit_primary_text_color',
        'homeTeamKit_secondaryColor': 'home_kit_secondary_color',
        'homeTeamKit_secondaryTextColor': 'home_kit_secondary_text_color',
        'awayTeam_id': 'away_team_id',
        'awayTeam_name': 'away_team_name',
        'awayTeamKit_primaryColor': 'away_kit_primary_color',
        'awayTeamKit_primaryTextColor': 'away_kit_primary_text_color',
        'awayTeamKit_secondaryColor': 'away_kit_secondary_color',
        'awayTeamKit_secondaryTextColor': 'away_kit_secondary_text_color',
    }

    skip_keys = [
        'complete',
        'gameEvents',
        'stadium',
        'videos',
        'period1',
        'period2',
        'rosters',
    ]
    metadata = {k: v for k, v in match.items() if k not in skip_keys}

    # Select which pitch to keep...
    if match.get('stadium'):
        metadata['stadium_name'] = match['stadium']['name']
        metadata['stadium_id'] = match['stadium']['id']

        # if match['stadium'].get('pitches'):
        #     for pitch in match['stadium']['pitches']:
        #         if pitch['startDate'] <= match['date'] <= pitch['endDate']:
        #             metadata.update({
        #                 'pitch_length': pitch['length'],
        #                 'pitch_width': pitch['width'],
        #             })
        #             break

    if match.get('videos'):
        video = match['videos'][0]
        metadata.update(
            {
                'video_fps': video['fps'],
            }
        )

    flat_metadata = _flatten_dict(metadata)

    return {
        new_name: flat_metadata.get(original_name)
        for original_name, new_name in metadata_field_map.items()
    }


def transform_player_metadata(rosters: list[dict[str, Any]]) -> list[dict[str, Any]]:
    player_field_map = {
        # Match context
        'game_id': 'match_id',
        'team_id': 'team_id',
        'team_name': 'team_name',
        'started': 'started',
        'shirtNumber': 'shirt_number',
        'positionGroupType': 'position_group_type',
        # Basic player info
        'player_id': 'player_id',
        'player_firstName': 'first_name',
        'player_lastName': 'last_name',
        'player_nickname': 'nickname',
        'player_dob': 'dob',
        'player_preferredFoot': 'preferred_foot',
        # Nationality
        'player_nationality_country': 'nationality',
        'player_secondNationality_country': 'second_nationality_country',
    }

    players = []

    for roster in rosters:
        all_fields = _flatten_dict(roster)

        player_data = {
            new_name: all_fields.get(original_name)
            for original_name, new_name in player_field_map.items()
        }

        players.append(player_data)

    return players


def _transform_game_event(event: dict[str, Any]) -> dict[str, Any]:
    """Process a game event, excluding specified fields and flattening the rest."""
    game_event_exclude_fields = {
        'advantageType',
        'defenderLocations',
        'offenderLocations',
        'videoMissing',
        'video',
        'videoAngleType',
        'subType',
        'heightType',
        'endType',
        'possessionEvents',
        'scoreValue',
    }

    filtered_event = {
        k: v for k, v in event.items() if k not in game_event_exclude_fields
    }

    return _flatten_dict(filtered_event, 'game_event')


def _transform_possession_event(event: dict[str, Any]) -> dict[str, Any]:
    """Process a possession event, excluding specified fields and flattening."""
    possession_event_exclude_fields = {
        'lastInGameEvent',
        'game',
        'defenders',
        'video',
        'period',
        'fouls',
    }

    filtered_event = {
        k: v
        for k, v in event.items()
        if k not in possession_event_exclude_fields and k not in POSSESSION_EVENT_TYPES
    }

    return _flatten_dict(filtered_event, 'possession_event')


def _transform_event_type(
    event_type: str, event_data: dict[str, Any]
) -> dict[str, Any]:
    """Process a specific event type (pass, shot, etc.) into flattened fields."""
    prefix = event_type.replace('Event', '')
    return _flatten_dict(event_data, prefix)


def _transform_foul(foul: dict[str, Any]) -> dict[str, Any]:
    """Process a foul inside a possession event into flattened fields."""
    return _flatten_dict(foul, 'foul')


def transform_events(game_events: list[dict[str, Any]]) -> list[dict[str, Any]]:
    """
    Transform PFF game events into flattened rows.

    Args:
        game_events: List of game event dictionaries

    Returns:
        List of dictionaries, each representing a flattened event row
    """
    transformed_rows = []

    for game_event in game_events:
        base_event_data = _transform_game_event(game_event)

        possession_events = game_event.get('possessionEvents', [])

        if not possession_events:
            transformed_rows.append(base_event_data)

        for possession_event in possession_events:
            possession_data = _transform_possession_event(possession_event)

            base_row_data = {**base_event_data, **possession_data}

            for event_type in POSSESSION_EVENT_TYPES:
                if event_data := possession_event.get(event_type):
                    event_row = base_row_data.copy()
                    event_row.update(_transform_event_type(event_type, event_data))
                    transformed_rows.append(event_row)

            for foul in possession_event.get('fouls', []):
                foul_row = base_row_data.copy()
                foul_row.update(_transform_foul(foul))
                transformed_rows.append(foul_row)

    mapped_rows = []

    for row in transformed_rows:
        mapped_row = {}

        for key, value in row.items():
            mapped_key = EVENTS_FIELD_MAP.get(key, key)
            mapped_row[mapped_key] = value

        mapped_rows.append(mapped_row)

    return mapped_rows



from typing import cast

import matplotlib.pyplot as plt
from IPython.display import HTML
from matplotlib.animation import FuncAnimation
from matplotlib.axes import Axes
from mplsoccer import Pitch

player_opts = {'s': 169, 'alpha': 0.8, 'linewidths': 1, 'marker': 'o'}
shirt_opts = {'ha': 'center', 'va': 'center', 'fontsize': 8, 'fontweight': 'bold'}
ball_opts = {'facecolor': 'black', 'edgecolor': 'lightyellow'}


def view_frame(frame: dict, metadata: list[dict] | None = None, ax: Axes | None = None):
    pitch = Pitch(pitch_type='impect', linewidth=1, line_alpha=0.8, goal_alpha=0.8)  # type: ignore
    if ax:
        pitch.draw(ax=ax)
    else:
        _, ax = pitch.draw()  # type: ignore
        ax = cast(Axes, ax)

    home_kit = metadata[0]['homeTeamKit'] if metadata else None
    away_kit = metadata[0]['awayTeamKit'] if metadata else None

    colors = {
        'home': (
            (home_kit['primaryColor'], home_kit['primaryTextColor'])
            if home_kit
            else ('darkred', 'white')
        ),
        'away': (
            (away_kit['primaryColor'], away_kit['primaryTextColor'])
            if away_kit
            else ('darkblue', 'white')
        ),
    }

    for side in ['home', 'away']:
        for player in frame[f'{side}PlayersSmoothed']:
            _plot_player(player, colors[side], ax)

    ball = frame['ballsSmoothed'] or (frame['balls'][0] if frame['balls'] else None)
    if ball:
        ball_x = ball['x']
        ball_y = ball['y']
        ax.scatter(ball_x, ball_y, **ball_opts, zorder=2)  # type: ignore

    return ax


def _plot_player(player: dict, colors: tuple, ax: Axes):
    x = player['x']
    y = player['y']
    shirt = player['jerseyNum']
    ax.scatter(x, y, facecolor=colors[0], edgecolor='black', **player_opts)
    if shirt:
        ax.text(x, y, str(shirt), c=colors[1], **shirt_opts, zorder=2)


def view_frame_sequence(
    frames: list[dict],
    metadata: list[dict] | None = None,
    *,
    ax: Axes | None = None,
    fps=30,
):
    pitch = Pitch(pitch_type='impect', linewidth=1, line_alpha=0.8, goal_alpha=0.8)  # type: ignore
    if ax:
        pitch.draw(ax=ax)
    else:
        fig, ax = pitch.draw()  # type: ignore
        ax = cast(Axes, ax)

    home_kit = metadata[0]['homeTeamKit'] if metadata else None
    away_kit = metadata[0]['awayTeamKit'] if metadata else None

    colors = {
        'home': (
            (home_kit['primaryColor'], home_kit['primaryTextColor'])
            if home_kit
            else ('darkred', 'white')
        ),
        'away': (
            (away_kit['primaryColor'], away_kit['primaryTextColor'])
            if away_kit
            else ('darkblue', 'white')
        ),
    }

    home_scatter = ax.scatter(
        [],
        [],
        facecolor=colors['home'][0],
        edgecolor='black',
        **player_opts,
    )
    away_scatter = ax.scatter(
        [],
        [],
        facecolor=colors['away'][0],
        edgecolor='black',
        **player_opts,
    )
    ball_scatter = ax.scatter([], [], **ball_opts, zorder=2)
    shirts = [ax.text(0, 0, '', **shirt_opts) for _ in range(22)]

    def update(frame_idx):
        nonlocal shirts

        frame = frames[frame_idx]

        home_positions = [[p['x'], p['y']] for p in frame['homePlayersSmoothed']]
        home_scatter.set_offsets(home_positions)

        away_positions = [[p['x'], p['y']] for p in frame['awayPlayersSmoothed']]
        away_scatter.set_offsets(away_positions)

        ball = frame['ballsSmoothed'] or (frame['balls'][0] if frame['balls'] else None)
        if ball:
            ball_x = ball['x']
            ball_y = ball['y']
            ball_position = [[ball_x, ball_y]]
            ball_scatter.set_offsets(ball_position)
            ball_scatter.set_visible(True)
        else:
            ball_scatter.set_visible(False)

        shirt_index = 0
        for pos, player in zip(
            home_positions, frame['homePlayersSmoothed'], strict=False
        ):
            x, y = pos[0], pos[1]
            shirts[shirt_index].set_position((x, y))
            shirts[shirt_index].set_text(str(player['jerseyNum']))
            shirts[shirt_index].set_color(colors['home'][1])
            shirts[shirt_index].set_visible(True)
            shirt_index += 1

        for pos, player in zip(
            away_positions, frame['awayPlayersSmoothed'], strict=False
        ):
            x, y = pos[0], pos[1]
            shirts[shirt_index].set_position((x, y))
            shirts[shirt_index].set_text(str(player['jerseyNum']))
            shirts[shirt_index].set_color(colors['away'][1])
            shirts[shirt_index].set_visible(True)
            shirt_index += 1

        for i in range(shirt_index, len(shirts)):
            shirts[i].set_visible(False)

        return (home_scatter, away_scatter, ball_scatter, *shirts)

    anim = FuncAnimation(
        fig,
        update,
        frames=len(frames),
        blit=True,
        interval=1000 / 30,
        repeat_delay=5000,
    )

    html_output = anim.to_jshtml()
    plt.close(fig)

    return HTML(html_output)



from datetime import datetime
from enum import Enum

from pydantic import BaseModel, ConfigDict


class PFFFrameJerseyConfidence(str, Enum):
    HIGH = 'HIGH'
    MEDIUM = 'MEDIUM'
    LOW = 'LOW'


class PFFFrameVisibility(str, Enum):
    VISIBLE = 'VISIBLE'
    ESTIMATED = 'ESTIMATED'


class PFFFrameGameEventType(str, Enum):
    FIRST_KICK_OFF = 'FIRSTKICKOFF'
    SECOND_KICK_OFF = 'SECONDKICKOFF'
    THIRD_KICK_OFF = 'THIRDKICKOFF'
    FOURTH_KICK_OFF = 'FOURTHKICKOFF'
    # FIRST_HALF_KICKOFF = '1KO'
    # SECOND_HALF_KICKOFF = '2KO'
    END_OF_HALF = 'END'
    PBC_IN_PLAY = 'G'
    PLAYER_ON = 'ON'
    PLAYER_OFF = 'OFF'
    ON_THE_BALL = 'OTB'
    OUT_OF_PLAY = 'OUT'
    SUB = 'SUB'
    VIDEO_MISSING = 'VID'
    CLOCK = 'CLK'  # TODO: Check if this is correct


class PFFFrameSetPieceType(str, Enum):
    CORNER = 'C'
    DROP_BALL = 'D'
    FREE_KICK = 'F'
    GOAL_KICK = 'G'
    KICK_OFF = 'K'
    PENALTY = 'P'
    THROW_IN = 'T'


class PFFFramePlayer(BaseModel):
    model_config = ConfigDict(extra='forbid')

    jerseyNum: int  # ignore
    confidence: PFFFrameJerseyConfidence
    visibility: PFFFrameVisibility
    x: float
    y: float
    speed: float | None = None


class PFFFrameBall(BaseModel):
    model_config = ConfigDict(extra='forbid')

    visibility: PFFFrameVisibility
    x: float | None
    y: float | None
    z: float | None


class PFFFrameEvent(BaseModel):
    model_config = ConfigDict(extra='forbid')

    game_id: int
    game_event_type: PFFFrameGameEventType
    formatted_game_clock: str | None
    player_id: int | None
    team_id: int | None
    setpiece_type: PFFFrameSetPieceType | None = None
    touches: int | None = None
    touches_in_box: int | None = None
    start_time: float
    end_time: float | None = None
    duration: float | None = None
    video_missing: bool | None = False
    inserted_at: datetime | None = None
    updated_at: datetime | None = None
    start_frame: int
    end_frame: int
    sequence: int | None = None
    home_ball: int | None = None
    competition_id: int | None = None
    season: str | None = None
    shirt_number: int | None
    position_group_type: str | None
    home_team: int | None
    player_name: str | None = None
    team_name: str | None = None
    video_url: str | None = None


class PFFFramePossessionEventType(str, Enum):
    BALL_CARRY = 'BC'
    CHALLENGE = 'CH'  # includes dribbles
    CLEARANCE = 'CL'
    CROSS = 'CR'
    PASS = 'PA'
    REBOUND = 'RE'
    SHOT = 'SH'


class PFFFramePossession(BaseModel):
    model_config = ConfigDict(extra='forbid')

    duration: float | None = None
    end_time: float | None = None
    formatted_game_clock: str | None
    game_event_id: int
    game_id: int
    inserted_at: datetime | None = None
    possession_event_type: PFFFramePossessionEventType
    start_time: float
    updated_at: datetime | None = None
    start_frame: int
    end_frame: int | None = None
    video_url: str | None = None


class PFFFrame(BaseModel):
    model_config = ConfigDict(
        populate_by_name=True, arbitrary_types_allowed=True, extra='forbid'
    )

    awayPlayers: list[PFFFramePlayer]
    awayPlayersSmoothed: list[PFFFramePlayer] | None
    balls: list[PFFFrameBall] | None
    ballsSmoothed: PFFFrameBall | None
    periodElapsedTime: float
    game_event: PFFFrameEvent | None
    game_event_id: int | None
    frameNum: int
    periodGameClockTime: float
    gameRefId: int | None
    generatedTime: datetime | None
    home_ball: int | None = None
    homePlayers: list[PFFFramePlayer]
    homePlayersSmoothed: list[PFFFramePlayer] | None
    period: int
    possession_event: PFFFramePossession | None
    possession_event_id: int | None
    sequence: int | None = None
    version: str | None
    videoTimeMs: float
    smoothedTime: datetime | None = None


class PFFFrames(BaseModel):
    frames: list[PFFFrame]



"""Provides a pitch for visualizing data on."""

from typing import cast

from matplotlib.axes import Axes
from matplotlib.figure import Figure
from mplsoccer import Pitch

from ..core.config import get_config_value

default_pitch_options = {
    'line_alpha': 0.8,
    'goal_alpha': 0.8,
}


def get_pitch(
    ax: Axes | None = None,
    *,
    pitch_length,
    pitch_width,
    **pitch_options,
) -> tuple[Pitch, Figure, Axes]:
    """Gets a pitch to visualize data on.

    Params:
        ax: matplotlib.axes.Axes | None
            A matplotlib.axes.Axes to draw on. If an ax is specified the pitch is drawn
            on an existing axis. Otherwise, a new figure is created
        pitch_length: float
            Horizontal size of the pitch
        pitch_width: float
            Vertical size of the pitch

    Returns:
        pitch: mplsoccer.Pitch
        fig: matplotlib.Figure
        ax: matplotlib.axes.Axes
    """
    pitch_values = get_config_value(section='pitch', default={})
    if pitch_length is None:
        pitch_length = pitch_values.get('length')
    if pitch_width is None:
        pitch_width = pitch_values.get('width')

    options = {**default_pitch_options, **pitch_options}
    pitch = Pitch(
        pitch_type='custom',
        pitch_length=pitch_length,
        pitch_width=pitch_width,
        **options,
    )

    if ax is None:
        fig, ax = pitch.draw(ax=None, figsize=(12, 8))  # type: ignore
        ax = cast(Axes, ax)
    else:
        pitch.draw(ax=ax)
        fig = ax.get_figure()

    fig = cast(Figure, fig)

    return pitch, fig, ax



from contextlib import contextmanager
from time import time


@contextmanager
def performance_logging(description: str, counter: int | None = None, logger=None):
    start = time()
    try:
        yield
    finally:
        took = (time() - start) * 1000
        extra = ''
        if counter is not None:
            extra = f' ({int(counter / took * 1000)}items/sec)'

        unit = 'ms'
        if took < 0.1:
            took *= 1000
            unit = 'us'

        msg = f'{description} took: {took:.2f}{unit} {extra}'
        if logger:
            logger.info(msg)
        else:
            print(msg)



"""Transform raw PFF tracking data into flattened structures for DataFrames."""

from enum import Enum
from itertools import chain
from typing import Any

from ..core.logging import get_logger

logger = get_logger(__name__)


class TransformFormat(str, Enum):
    """Format options for the transformed output."""

    WIDE = 'wide'  # One row per frame with player data as columns
    LONG = 'long'  # Multiple rows per frame (one per player)


def _flatten_dict(d: dict[str, Any], prefix: str = '') -> dict[str, Any]:
    """
    Recursively flatten a nested dictionary with prefixed keys.

    Args:
        d: Dictionary to flatten
        prefix: Prefix to add to keys (with trailing underscore if non-empty)

    Returns:
        Flattened dictionary where nested keys are combined with their parent keys
    """
    result = {}

    try:
        for k, v in d.items():
            key = f'{prefix}_{k}' if prefix else k

            if isinstance(v, dict):
                result.update(_flatten_dict(v, key))
            else:
                result[key] = v
    except AttributeError:
        logger.error(f'Error flattening dictionary: {d}')
        raise

    return result


def _transform_wide(frame: dict[str, Any]) -> dict[str, Any]:
    """
    Transform a single frame dictionary to a flat structure.

    Args:
        frame: A dictionary containing PFF tracking data for one frame

    Returns:
        A flat dictionary containing all frame data
    """
    # fmt: off
    skip_keys = [
        'homePlayers', 'awayPlayers', 'balls',
        'homePlayersSmoothed', 'awayPlayersSmoothed', 'ballsSmoothed' # kalman filtered,
        'team'
    ]
    # fmt: on

    result = {
        k: v
        for k, v in frame.items()
        if k not in skip_keys and not isinstance(v, dict) and not isinstance(v, list)
    }

    for key in ['game_event', 'possession_event']:
        if frame.get(key):
            result.update(_flatten_dict(frame[key], key))

    player_lists = [
        # ('home', 'homePlayers'),
        # ('away', 'awayPlayers'),
        ('home', 'homePlayersSmoothed'),
        ('away', 'awayPlayersSmoothed'),
    ]

    for team, key in player_lists:
        result[f'{team}_count'] = 0
        if key in frame and frame[key] is not None:
            players = frame[key]
            result[f'{team}_count'] = len(players)

            for player in players:
                if 'jerseyNum' in player:
                    jersey_num = player['jerseyNum']
                    for attr, value in player.items():
                        result[f'{team}_{jersey_num}_{attr}'] = value
                else:
                    logger.warning(f'Found player without jerseyNum in {key}')

    # ballsSmoothed doesn't need to be smoothed, just ball, we'll drop ballsSmoothed
    if 'balls' in frame and frame.get('balls'):
        ball = frame['balls'][0]
        result.update(_flatten_dict(ball, 'ball'))

    return result


def _transform_long(frame: dict[str, Any]) -> list[dict[str, Any]]:
    """
    Transform a single frame dictionary to a long format structure with one row per
    player.

    In this format, each player becomes a separate row with common frame data repeated.
    Ball data and events are included in each row.

    Args:
        frame: A dictionary containing PFF tracking data for one frame

    Returns:
        A list of dictionaries, one per player
    """
    # fmt: off
    skip_keys = [
        'homePlayers', 'awayPlayers', 'balls',
        'homePlayersSmoothed', 'awayPlayersSmoothed', 'ballsSmoothed'
    ]
    # fmt: on

    common_fields = {}
    for k, v in frame.items():
        if k not in skip_keys and not isinstance(v, dict) and not isinstance(v, list):
            common_fields[k] = v

    if 'balls' in frame and frame.get('balls'):
        ball = frame['balls'][0]
        common_fields.update(_flatten_dict(ball, 'ball'))

    for key in ['game_event', 'possession_event']:
        if frame.get(key):
            common_fields.update(_flatten_dict(frame[key], key))

    results = []

    player_lists = [
        # ('home', 'homePlayers'),
        # ('away', 'awayPlayers'),
        ('home', 'homePlayersSmoothed'),
        ('away', 'awayPlayersSmoothed'),
    ]

    for team, key in player_lists:
        if key in frame and frame[key] is not None:
            players = frame[key]
            for player in players:
                player_row = common_fields.copy()
                player_row['team'] = team

                for attr, value in player.items():
                    player_row[f'player_{attr}'] = value

                results.append(player_row)

    return results


def transform_pff_tracks(
    frames: list[dict[str, Any]],
    transform_format: TransformFormat = TransformFormat.WIDE,
) -> list[dict[str, Any]]:
    """
    Transform a list of frame dictionaries into flat structures.

    Args:
        frames: A list of dictionaries containing PFF tracking data
        transform_format: Format for the output data (wide or long)

    Returns:
        In WIDE format: A list of flattened dictionaries (one per frame)
        In LONG format: A list of lists of dictionaries (multiple rows per frame)
    """
    if transform_format not in TransformFormat:
        raise ValueError(f'Invalid transform format: {transform_format}')

    if transform_format == TransformFormat.WIDE:
        return [_transform_wide(frame) for frame in frames]
    else:
        return list(chain.from_iterable(_transform_long(frame) for frame in frames))



"""
Configuration settings for gandula. Sets up logging, paths, and validation settings.
"""

import importlib.resources
import os
import sys
from pathlib import Path
from typing import Any

import yaml


def load_yaml_config() -> dict[str, Any]:
    """
    Load configuration from YAML file.

    Returns:
        Configuration dictionary loaded from the YAML file
    """
    try:
        config_text = (
            importlib.resources.files('gandula').joinpath('config.yaml').read_text()
        )
        config = yaml.safe_load(config_text)

        if not config:
            raise ValueError('Config file is empty or failed to load')
    except (yaml.YAMLError, OSError, ModuleNotFoundError, ImportError, ValueError) as e:
        print(f'Error loading configuration from package: {e}', file=sys.stderr)
        raise

    return config


def load_user_config() -> dict[str, Any]:
    """
    Load user configuration from ~/.config/gandula/config.yaml if it exists.

    Returns:
        User configuration dictionary
    """
    explicit_path = os.environ.get('GANDULA_CONFIG')
    if explicit_path:
        config_path = Path(explicit_path)
        if config_path.exists():
            try:
                with config_path.open() as f:
                    return yaml.safe_load(f) or {}
            except (yaml.YAMLError, OSError) as e:
                print(
                    f'Error loading explicit config from {config_path}: {e}',
                    file=sys.stderr,
                )

    # Check standard user config locations
    user_paths = [
        Path.home() / '.config' / 'gandula' / 'config.yaml',
        Path.home() / '.gandula.config.yaml',
    ]

    for path in user_paths:
        if path.exists():
            try:
                with path.open() as f:
                    user_config = yaml.safe_load(f)
                    if user_config:
                        print(f'Loaded user configuration from {path}', file=sys.stderr)
                        return user_config
            except (yaml.YAMLError, OSError) as e:
                print(
                    f'Error loading user configuration from {path}: {e}',
                    file=sys.stderr,
                )
                raise

    return {}


def merge_configs(
    base_config: dict[str, Any], override_config: dict[str, Any]
) -> dict[str, Any]:
    """
    Deep merge two configuration dictionaries.

    Args:
        base_config: Base configuration
        override_config: Configuration that takes precedence

    Returns:
        Merged configuration dictionary
    """
    merged = base_config.copy()

    for section, values in override_config.items():
        if (
            section in merged
            and isinstance(merged[section], dict)
            and isinstance(values, dict)
        ):
            merged[section] = merge_configs(merged[section], values)
        else:
            merged[section] = values

    return merged


def get_config_value(section: str, key: str | None = None, default: Any = None) -> Any:
    """
    Get a configuration value from the loaded configuration.

    Args:
        section: Configuration section
        key: Configuration key
        default: Default value if not found

    Returns:
        Configuration value or default
    """
    if key is None:
        return CONFIG.get(section, default)

    try:
        return CONFIG[section][key]
    except (KeyError, TypeError):
        return default


def configure(config_updates: dict[str, Any]) -> None:
    """
    Updates the configuration at runtime.

    Allows dynamic modification of the configuration after initial setup.

    Args:
        config_updates: A dictionary containing configuration key-value pairs to update.
            The dictionary should follow the structure of the existing configuration,
            with nested updates supported for complex configuration structures.

    Raises:
        ValueError: If the provided configuration updates are invalid or cannot be
            applied.

    Examples:
        Update a specific data directory path:
        ```python
            import gandula.config

            # Change the data directory
            gandula.config.configure({"paths": {"data_dir": "/new/path"}})

            # Update multiple configuration options
            gandula.config.configure({
                "paths": {
                    "data_dir": "/updated/data/path",
                    "output_dir": "/new/output/path"
                },
                "logging": {
                    "level": "DEBUG"
                }
            })
        ```
    """
    global CONFIG

    CONFIG = merge_configs(CONFIG, config_updates)

    if 'logging' in config_updates:
        # to avoid circular imports
        from .logging import configure_logging

        configure_logging()


def get_data_dir() -> Path:
    """
    Get the data directory path.

    Returns:
        Path to the data directory
    """
    paths_config = CONFIG.get('paths', {})
    data_dir = paths_config.get('data_dir', os.path.join(os.getcwd(), 'data'))

    path = Path(data_dir)

    if not path.exists():
        raise FileNotFoundError(f'Data directory does not exist: {path}')

    return path


def get_temp_dir() -> Path:
    """
    Get the temporary directory path.

    Returns:
        Path to the temporary directory
    """
    import tempfile

    paths_config = CONFIG.get('paths', {})
    temp_dir = paths_config.get('temp_dir')

    if temp_dir:
        path = Path(temp_dir)

        if not path.exists():
            path.mkdir(parents=True, exist_ok=True)
        return path

    return Path(tempfile.gettempdir())


package_config = load_yaml_config()
user_config = load_user_config()
CONFIG = merge_configs(package_config, user_config)



from datetime import timedelta
from itertools import chain, zip_longest
from typing import NamedTuple

from kloppy.domain import (
    DatasetFlag,
    EventDataset,
    Ground,
    Metadata,
    Orientation,
    Period,
    Player,
    Provider,
    Team,
)
from kloppy.exceptions import DeserializationError
from kloppy.infra.serializers.event.deserializer import EventDataDeserializer
from kloppy.utils import performance_logging

from ..core.logging import get_logger
from . import pff_event_specification as PFF
from .pff_event_specification import position_types_mapping

logger = get_logger(__name__)


class PFFInputs(NamedTuple):
    event_data: dict


# TODO: Move all fixes (periods, teams, etc) BEFORE deserializing
class PFFDeserializer(EventDataDeserializer[PFFInputs]):
    def provider(self):
        return Provider.OTHER

    def deserialize(self, inputs: PFFInputs, additional_metadata) -> EventDataset:
        self.transformer = self.get_transformer(
            pitch_length=105.0, pitch_width=68.0, provider=Provider.KLOPPY
        )

        with performance_logging('load data', logger=logger):
            raw_game, raw_events, data_version = self.load_data(inputs)
        logger.info(f'determined version {data_version} from event data')

        with performance_logging('parse teams and players', logger=logger):
            teams = self.create_teams_and_players(raw_game)

        with performance_logging('parse periods', logger=logger):
            periods = self.create_periods(raw_events)

        with performance_logging('parse events', logger=logger):
            events = []
            for raw_event in chain.from_iterable(raw_events.values()):
                new_events = (
                    raw_event.set_version(data_version)
                    .set_refs(periods, teams, raw_events)
                    .deserialize(self.event_factory)
                )
                for event in new_events:
                    if self.should_include_event(event):
                        # event = self.transformer.transform_event(event)
                        events.append(event)

        metadata = Metadata(
            teams=teams,
            periods=periods,
            pitch_dimensions=self.transformer.get_to_coordinate_system().pitch_dimensions,
            frame_rate=None,
            orientation=Orientation.ACTION_EXECUTING_TEAM,
            flags=DatasetFlag.BALL_OWNING_TEAM | DatasetFlag.BALL_STATE,
            score=None,
            provider=Provider.OTHER,
            coordinate_system=self.transformer.get_to_coordinate_system(),
            **additional_metadata,
        )

        return EventDataset(metadata=metadata, records=events)

    def load_data(self, inputs: PFFInputs):
        raw_events = {}
        version = ''

        raw_game = inputs.event_data

        with performance_logging('sort events', logger=logger):
            raw_game['gameEvents'] = sorted(
                raw_game['gameEvents'], key=lambda x: x['startTime']
            )

        for event in raw_game['gameEvents']:
            raw_events[event['id']] = PFF.event_decoder(event)
            if not version:
                version = event.get('version')

        version = PFF.Version(version=version)

        return raw_game, raw_events, version

    def create_teams_and_players(self, raw_game):
        rosters = raw_game.get('rosters')

        home_team = raw_game['homeTeam']
        away_team = raw_game['awayTeam']

        home_team['players'] = [
            player for player in rosters if player['team']['id'] == home_team.get('id')
        ]
        away_team['players'] = [
            player for player in rosters if player['team']['id'] == away_team.get('id')
        ]

        player_positions = {
            player['player']['id']: position_types_mapping[player['positionGroupType']]
            for player in rosters
        }

        def create_team(team_players, ground_type):
            team = Team(
                team_id=team_players['id'],
                name=team_players['name'],
                ground=ground_type,
            )
            team.players = [
                Player(
                    player_id=str(player['player']['id']),
                    team=team,
                    first_name=player['player']['firstName'],
                    last_name=player['player']['lastName'],
                    name=player['player']['nickname'],
                    jersey_no=int(player['shirtNumber']),
                    starting=player['started'],
                    starting_position=player_positions.get(player['player']['id']),
                )
                for player in team_players['players']
            ]
            return team

        home_team = create_team(home_team, Ground.HOME)
        away_team = create_team(away_team, Ground.AWAY)

        return [home_team, away_team]

    def create_periods(self, raw_events) -> list[Period]:
        half_start_events = {}
        half_end_events = {}

        for event in chain.from_iterable(raw_events.values()):
            event_type = PFF.EVENT_TYPE(event.raw_event['gameEventType'])
            period = event.raw_event['period']

            # There is a bug where SECOND_HALF_KICKOFF period is 1
            # easier to control directly and fix this
            fix_kickoff_period = {
                PFF.EVENT_TYPE.FIRST_HALF_KICKOFF: 1,
                PFF.EVENT_TYPE.SECOND_HALF_KICKOFF: 2,
                PFF.EVENT_TYPE.EXTRA_1_KICKOFF: 3,
                PFF.EVENT_TYPE.EXTRA_2_KICKOFF: 4,
            }

            if event_type in [
                PFF.EVENT_TYPE.FIRST_HALF_KICKOFF,
                PFF.EVENT_TYPE.SECOND_HALF_KICKOFF,
                PFF.EVENT_TYPE.EXTRA_1_KICKOFF,
                PFF.EVENT_TYPE.EXTRA_2_KICKOFF,
            ]:
                period = fix_kickoff_period[event_type]
                event.raw_event['period'] = period
                half_start_events[int(period)] = event.raw_event
            elif event_type == PFF.EVENT_TYPE.END_OF_HALF:
                half_end_events[int(period)] = event.raw_event

        periods = []
        for start_event, end_event in zip_longest(
            half_start_events.values(), half_end_events.values()
        ):
            if start_event is None or end_event is None:
                raise DeserializationError(
                    'Failed to determine start and end time of periods.'
                )

            period = Period(
                id=int(start_event['period']),
                start_timestamp=timedelta(seconds=start_event['startTime']),
                end_timestamp=timedelta(seconds=end_event['startTime']),
            )

            periods.append(period)

        return periods
